# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Create a couple of different mesh cases and test the performance."""

import numpy as np
import pytest

from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.input_file import InputFile
from beamme.four_c.material import MaterialReissner
from beamme.four_c.model_importer import import_four_c_model
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.utils.environment import cubitpy_is_available
from beamme.utils.nodes import find_close_nodes

if cubitpy_is_available():
    from cubitpy import CubitPy, cupy


@pytest.fixture(scope="module")
def shared_tmp_path(tmp_path_factory):
    """Create a temporary path for shared use in performance tests."""
    return tmp_path_factory.mktemp("performance_tests")


def create_solid_block(cubit, file_path, nx, ny, nz):
    """Create a solid block (1 x 1 x 1) with (nx * ny * nz) elements."""

    # Create brick.
    brick = cubit.brick(1)

    # Set mesh parameters.
    mesh_size = [
        [nx, [2, 4, 6, 8]],
        [ny, [1, 3, 5, 7]],
        [nz, [9, 10, 11, 12]],
    ]
    for [n_el, curves] in mesh_size:
        for i in curves:
            cubit.set_line_interval(brick.curves()[i - 1], n_el)
    brick.volumes()[0].mesh()

    # Add block and sets.
    cubit.add_element_type(
        brick.volumes()[0],
        cupy.element_type.hex8,
        name="brick",
        material={"MAT": 1},
        bc_description={"KINEM": "nonlinear"},
    )
    counter = 0
    for item in brick.vertices():
        cubit.add_node_set(
            item,
            name="node_set_" + str(counter),
            bc_type=cupy.bc_type.neumann,
            bc_description={
                "NUMDOF": 3,
                "ONOFF": [1, 1, 1],
                "VAL": [3.0, 3.0, 0],
                "FUNCT": [1, 2, 0],
            },
        )
        counter += 1
    for item in brick.curves():
        cubit.add_node_set(
            item,
            name="node_set_" + str(counter),
            bc_type=cupy.bc_type.dirichlet,
            bc_description={
                "NUMDOF": 3,
                "ONOFF": [1, 1, 1],
                "VAL": [3.0, 3.0, 0],
                "FUNCT": [1, 2, 0],
            },
        )
        counter += 1
    for item in brick.surfaces():
        cubit.add_node_set(
            item,
            name="node_set_" + str(counter),
            bc_type=cupy.bc_type.neumann,
            bc_description={
                "NUMDOF": 3,
                "ONOFF": [1, 1, 1],
                "VAL": [3.0, 3.0, 0],
                "FUNCT": [1, 2, 0],
            },
        )
        counter += 1
    for item in brick.volumes():
        cubit.add_node_set(
            item,
            name="node_set_" + str(counter),
            bc_type=cupy.bc_type.neumann,
            bc_description={
                "NUMDOF": 3,
                "ONOFF": [1, 1, 1],
                "VAL": [3.0, 3.0, 0],
                "FUNCT": [1, 2, 0],
            },
        )
        counter += 1

    # Set the material.
    cubit.fourc_input["MATERIALS"] = [
        {"MAT": 1, "MAT_Struct_StVenantKirchhoff": {"DENS": 1, "NUE": 0.3, "YOUNG": 2}}
    ]

    # Export mesh
    cubit.dump(file_path)


def create_beam_mesh(n_x, n_y, n_z, n_el):
    """Create a beam grid on the domain (1 x 1 x 1) with (nx * ny * nz) "grid
    cells"."""

    mesh = Mesh()
    material = MaterialReissner(
        radius=0.25 / np.max([n_x, n_y, n_z]), youngs_modulus=1.0
    )

    for i_x in range(n_x + 1):
        for i_y in range(n_y + 1):
            create_beam_mesh_line(
                mesh,
                Beam3rHerm2Line3,
                material,
                [i_x / n_x, i_y / n_y, 0],
                [i_x / n_x, i_y / n_y, 1],
                n_el=n_z * n_el,
            )
    for i_y in range(n_y + 1):
        for i_z in range(n_z + 1):
            create_beam_mesh_line(
                mesh,
                Beam3rHerm2Line3,
                material,
                [0, i_y / n_y, i_z / n_z],
                [1, i_y / n_y, i_z / n_z],
                n_el=n_x * n_el,
            )
    for i_z in range(n_z + 1):
        for i_x in range(n_x + 1):
            create_beam_mesh_line(
                mesh,
                Beam3rHerm2Line3,
                material,
                [i_x / n_x, 0, i_z / n_z],
                [i_x / n_x, 1, i_z / n_z],
                n_el=n_y * n_el,
            )
    return mesh


@pytest.mark.performance
def test_performance_beamme_cubitpy_create_solid(
    evaluate_execution_time, shared_tmp_path
):
    """Test the performance of creating a solid block using CubitPy.

    The version of Cubit we use in testing only allows for 50,000, so we
    create a mesh with exactly that.
    """

    cubit = CubitPy()

    evaluate_execution_time(
        "CubitPy: Create half solid block",
        create_solid_block,
        kwargs={
            "cubit": cubit,
            "file_path": shared_tmp_path / "performance_testing_solid_half.4C.yaml",
            "nx": 100,
            "ny": 50,
            "nz": 10,
        },
        expected_time=2.7,
    )


@pytest.mark.performance
def test_performance_beamme_double_solid_block(
    evaluate_execution_time, shared_tmp_path
):
    """The version of Cubit we use in testing only allows for 50,000 elements,
    our goal is 100,000 so we double the block with 50,000 elements here."""

    def double_block():
        """Load the block with 50,000 elements and double the mesh."""
        input_file, mesh = import_four_c_model(
            shared_tmp_path / "performance_testing_solid_half.4C.yaml",
            convert_input_to_mesh=True,
        )
        mesh_translated = mesh.copy()
        mesh_translated.translate([1, 0, 0])
        input_file.add(mesh)
        input_file.add(mesh_translated)
        input_file.dump(
            shared_tmp_path / "performance_testing_solid.4C.yaml", validate=False
        )

    evaluate_execution_time(
        "BeamMe: Load and double solid element block from cubit",
        double_block,
        expected_time=10.5,
    )


@pytest.mark.parametrize(
    ("log_name", "full_import", "expected_time"),
    [
        ("BeamMe: Load solid mesh (no full import)", False, 2.5),
        ("BeamMe: Load solid mesh (full import)", True, 3.5),
    ],
)
@pytest.mark.performance
def test_performance_beamme_load_solid(
    evaluate_execution_time, shared_tmp_path, log_name, full_import, expected_time
):
    """Test the performance of loading a solid mesh."""

    evaluate_execution_time(
        log_name,
        import_four_c_model,
        kwargs={
            "input_file_path": shared_tmp_path / "performance_testing_solid.4C.yaml",
            "convert_input_to_mesh": full_import,
        },
        expected_time=expected_time,
    )


@pytest.mark.performance
def test_performance_beamme_create_beams(evaluate_execution_time, cache_data):
    """Test the performance of creating a large beam mesh."""

    # store mesh in cache for upcoming tests
    cache_data.mesh = evaluate_execution_time(
        "BeamMe: Create large beam mesh",
        create_beam_mesh,
        kwargs={
            "n_x": 40,
            "n_y": 40,
            "n_z": 10,
            "n_el": 2,
        },
        expected_time=5.2,
    )


@pytest.mark.performance
def test_performance_beamme_rotate(evaluate_execution_time, cache_data):
    """Test the performance of rotating a large beam mesh."""

    evaluate_execution_time(
        "BeamMe: Rotate large beam mesh",
        cache_data.mesh.rotate,
        kwargs={"rotation": Rotation([1, 1, 0], np.pi / 3)},
        expected_time=0.5,
    )


@pytest.mark.performance
def test_performance_beamme_translate(evaluate_execution_time, cache_data):
    """Test the performance of translating a large beam mesh."""

    evaluate_execution_time(
        "BeamMe: Translate large beam mesh",
        cache_data.mesh.translate,
        kwargs={"vector": [0.5, 0, 0]},
        expected_time=0.25,
    )


@pytest.mark.performance
def test_performance_beamme_reflect(evaluate_execution_time, cache_data):
    """Test the performance of reflecting a large beam mesh."""

    evaluate_execution_time(
        "BeamMe: Reflect large beam mesh",
        cache_data.mesh.reflect,
        kwargs={"normal_vector": [0.5, 0.4, 0.1]},
        expected_time=0.5,
    )


@pytest.mark.performance
def test_performance_beamme_mespy_wrap_around_cylinder(
    evaluate_execution_time, cache_data
):
    """Test the performance of wrapping a large beam mesh around a cylinder."""

    evaluate_execution_time(
        "BeamMe: Wrap large beam mesh around cylinder",
        cache_data.mesh.wrap_around_cylinder,
        kwargs={"radius": 1.0},
        expected_time=1.75,
    )


@pytest.mark.performance
def test_performance_beamme_wrap_around_cylinder_without_check(
    evaluate_execution_time, cache_data
):
    """Test the performance of wrapping a large beam mesh around a cylinder
    without checking for advanced warnings."""

    evaluate_execution_time(
        "BeamMe: Wrap large beam mesh around cylinder without check",
        cache_data.mesh.wrap_around_cylinder,
        kwargs={"radius": 1.0, "advanced_warning": False},
        expected_time=0.5,
    )


@pytest.mark.performance
def test_performance_beamme_find_close_nodes(evaluate_execution_time, cache_data):
    """Test the performance of finding close nodes in a large beam mesh."""

    evaluate_execution_time(
        "BeamMe: Find close nodes in large beam mesh",
        find_close_nodes,
        kwargs={"nodes": cache_data.mesh.nodes},
        expected_time=0.4,
    )


@pytest.mark.performance
def test_performance_beamme_add_mesh_to_input_file(evaluate_execution_time, cache_data):
    """Test the performance of adding a mesh to an input file."""

    input_file = InputFile()

    evaluate_execution_time(
        "BeamMe: Add large beam mesh to input file",
        input_file.add,
        kwargs={"object_to_add": cache_data.mesh},
        expected_time=10.0,
    )

    cache_data.input_file = input_file


@pytest.mark.performance
def test_performance_beamme_dump_input_file(
    evaluate_execution_time,
    cache_data,
    tmp_path,
):
    """Test the performance of dumping an input file with a large beam mesh."""

    evaluate_execution_time(
        "BeamMe: Dump input file with large beam mesh",
        cache_data.input_file.dump,
        kwargs={
            "input_file_path": tmp_path / "performance_testing_beam.4C.yaml",
            "validate_sections_only": True,
        },
        expected_time=5.5,
    )


@pytest.mark.performance
def test_performance_beamme_write_vtk(evaluate_execution_time, tmp_path, cache_data):
    """Test the performance of writing a beam mesh to VTK format."""

    # use a smaller mesh for testing vtk output performance
    cache_data.mesh = create_beam_mesh(n_x=20, n_y=20, n_z=10, n_el=2)

    evaluate_execution_time(
        "BeamMe: Write beam mesh to VTK",
        cache_data.mesh.write_vtk,
        kwargs={
            "output_name": "performance_testing_beam",
            "output_directory": tmp_path,
            "beam_centerline_visualization_segments": 1,
        },
        expected_time=4.0,
    )


@pytest.mark.performance
def test_performance_beamme_write_vtk_smooth(
    evaluate_execution_time, tmp_path, cache_data
):
    """Test the performance of writing a beam mesh to VTK format with more
    segments."""

    evaluate_execution_time(
        "BeamMe: Write beam mesh to VTK with more segments",
        cache_data.mesh.write_vtk,
        kwargs={
            "output_name": "performance_testing_beam",
            "output_directory": tmp_path,
            "beam_centerline_visualization_segments": 5,
        },
        expected_time=8.1,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Integration tests for the beam interaction condition functionality."""

import pytest

from beamme.core.conf import bme
from beamme.core.mesh import Mesh
from beamme.four_c.beam_interaction_conditions import add_beam_interaction_condition
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.model_importer import import_four_c_model
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_integration_four_c_beam_interaction_conditions_beam_to_beam_contact(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the beam-to-beam contact boundary conditions."""

    # Create the mesh.
    mesh = Mesh()

    # Create Material.
    mat = get_default_test_beam_material(material_type="reissner")

    # Create a beam in x-axis.
    beam_x = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0, 0, 0],
        [1, 0, 0],
        n_el=2,
    )

    # Create a second beam in y-axis.
    beam_y = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0, 0, 0.5],
        [1, 0, 0.5],
        n_el=2,
    )

    # Add the beam-to-beam contact condition.
    add_beam_interaction_condition(
        mesh, beam_x["line"], beam_y["line"], bme.bc.beam_to_beam_contact
    )

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_four_c_beam_interaction_conditions_beam_to_solid(
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that the automatic ID creation for beam-to-solid conditions
    works."""

    # Load a solid
    _, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_block"
        ),
        convert_input_to_mesh=True,
    )

    # The yaml file already contains the beam-to-solid boundary conditions
    # for the solid. We don't need them in this test case, as we want to
    # create them again. Thus, we have to delete them here.
    mesh.boundary_conditions[
        (bme.bc.beam_to_solid_volume_meshtying, bme.geo.volume)
    ].clear()
    mesh.boundary_conditions[
        (bme.bc.beam_to_solid_surface_meshtying, bme.geo.surface)
    ].clear()

    # Get the geometry set objects representing the geometry from the cubit
    # file.
    surface_set = mesh.geometry_sets[bme.geo.surface][0]
    volume_set = mesh.geometry_sets[bme.geo.volume][0]

    # Add the beam
    material = get_default_test_beam_material(material_type="reissner")
    beam_set_1 = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, material, [0, 0, 0], [0, 0, 1], n_el=1
    )
    beam_set_2 = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, material, [0, 1, 0], [0, 1, 1], n_el=2
    )
    add_beam_interaction_condition(
        mesh,
        volume_set,
        beam_set_1["line"],
        bme.bc.beam_to_solid_volume_meshtying,
    )
    add_beam_interaction_condition(
        mesh,
        volume_set,
        beam_set_2["line"],
        bme.bc.beam_to_solid_volume_meshtying,
    )
    add_beam_interaction_condition(
        mesh,
        surface_set,
        beam_set_2["line"],
        bme.bc.beam_to_solid_surface_meshtying,
    )
    add_beam_interaction_condition(
        mesh,
        surface_set,
        beam_set_1["line"],
        bme.bc.beam_to_solid_surface_meshtying,
    )

    assert_results_close(get_corresponding_reference_file_path(), mesh)

    # If we try to add this the IDs won't match, because the next volume ID for
    # beam-to-surface coupling should be 0 (this one does not make sense, but
    # this is checked in a later test) and the next line ID for beam-to-surface
    # coupling is 2 (there are already two of these conditions).
    with pytest.raises(ValueError):
        add_beam_interaction_condition(
            mesh,
            volume_set,
            beam_set_1["line"],
            bme.bc.beam_to_solid_surface_meshtying,
        )

    # If we add a wrong geometries to the mesh, the creation of the input file
    # should fail, because there is no beam-to-surface contact section that
    # contains a volume set.
    with pytest.raises(KeyError):
        add_beam_interaction_condition(
            mesh,
            volume_set,
            beam_set_1["line"],
            bme.bc.beam_to_solid_surface_contact,
        )
        assert_results_close(get_corresponding_reference_file_path(), mesh)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Helper functions for the geometric search module."""

from importlib.util import find_spec as _find_spec


def arborx_is_available() -> bool:
    """Check if ArborX is available.

    Returns:
        True if ArborX is installed, False otherwise
    """

    if _find_spec("beamme.geometric_search.arborx_lib") is None:
        return False
    return True


def cython_is_available() -> bool:
    """Check if Cython is available.

    Returns:
        True if Cython is installed, False otherwise
    """

    if _find_spec("beamme.geometric_search.cython_lib") is None:
        return False
    return True



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Find unique points in a point cloud, i.e., points that are within a certain
tolerance of each other will be considered as unique."""

from enum import Enum as _Enum
from enum import auto as _auto

from beamme.geometric_search.scipy import (
    find_close_points_scipy as _find_close_points_scipy,
)
from beamme.geometric_search.utils import arborx_is_available as _arborx_is_available
from beamme.geometric_search.utils import cython_is_available as _cython_is_available

if _cython_is_available():
    from beamme.geometric_search.cython import (
        find_close_points_brute_force_cython as _find_close_points_brute_force_cython,
    )

if _arborx_is_available():
    from beamme.geometric_search.arborx import (
        find_close_points_arborx as _find_close_points_arborx,
    )


class FindClosePointAlgorithm(_Enum):
    """Enum for different find_close_point algorithms."""

    kd_tree_scipy = _auto()
    brute_force_cython = _auto()
    boundary_volume_hierarchy_arborx = _auto()


def point_partners_to_unique_indices(point_partners):
    """Convert the partner indices to lists that can be used for converting
    between the full and unique coordinates.

    Returns
    ----
    unique_indices: list(int)
        Indices that result in the unique point coordinate array.
    inverse_indices: list(int)
        Indices of the unique array that can be used to reconstruct of the original points coordinates.
    """

    unique_indices = []
    inverse_indices = [-1 for i in range(len(point_partners))]
    partner_id_to_unique_map = {}
    i_partner = 0
    for i_point, partner_index in enumerate(point_partners):
        if partner_index == -1:
            # This point does not have any partners, i.e., it is already a unique point.
            unique_indices.append(i_point)
            my_inverse_index = len(unique_indices) - 1
        elif partner_index == i_partner:
            # This point has partners and this is the first time that this partner index
            # appears in the input list.
            unique_indices.append(i_point)
            my_inverse_index = len(unique_indices) - 1
            partner_id_to_unique_map[partner_index] = my_inverse_index
            i_partner += 1
        elif partner_index < i_partner:
            # This point has partners and the partner index has been previously found.
            my_inverse_index = partner_id_to_unique_map[partner_index]
        else:
            raise ValueError(
                "This should not happen, as the partners should be provided in order"
            )
        inverse_indices[i_point] = my_inverse_index

    return unique_indices, inverse_indices


def point_partners_to_partner_indices(point_partners, n_partners):
    """Convert the partner indices for each point to a list of lists with the
    indices for all partners."""
    partner_indices = [[] for i in range(n_partners)]
    for i, partner_index in enumerate(point_partners):
        if partner_index != -1:
            partner_indices[partner_index].append(i)
    return partner_indices


def partner_indices_to_point_partners(partner_indices, n_points):
    """Convert the list of lists with the indices for all partners to the
    partner indices for each point."""
    point_partners = [-1 for _i in range(n_points)]
    for i_partner, partners in enumerate(partner_indices):
        for index in partners:
            point_partners[index] = i_partner
    return point_partners, len(partner_indices)


def find_close_points(point_coordinates, *, algorithm=None, tol=1e-8, **kwargs):
    """Find unique points in a point cloud, i.e., points that are within a
    certain tolerance of each other will be considered as unique.

    Args
    ----
    point_coordinates: _np.array(n_points x n_dim)
        Point coordinates that are checked for partners. The number of spatial dimensions
        does not have to be equal to 3.
    algorithm: FindClosePointAlgorithm
        Type of geometric search algorithm that should be used.
    n_bins: list(int)
        Number of bins in the first three dimensions.
    tol: float
        If the absolute distance between two points is smaller than tol, they
        are considered to be equal, i.e., tol is the hyper sphere radius that
        the point coordinates have to be within, to be identified as overlapping.

    Return
    ----
    has_partner: array(int)
        An array with integers, marking the partner index of each point. A partner
        index of -1 means the node does not have a partner.
    partner: int
        Largest partner index.
    """

    n_points = len(point_coordinates)

    if algorithm is None:
        # Decide which algorithm to use
        if n_points < 200 and _cython_is_available():
            # For around 200 points the brute force cython algorithm is the fastest one
            algorithm = FindClosePointAlgorithm.brute_force_cython
        elif _arborx_is_available():
            # For general problems with n_points > 200 the ArborX implementation is the fastest one
            algorithm = FindClosePointAlgorithm.boundary_volume_hierarchy_arborx
        else:
            # The scipy implementation is slower than ArborX by a factor of about 2, but is scales
            # the same
            algorithm = FindClosePointAlgorithm.kd_tree_scipy

    # Get list of closest pairs
    if algorithm is FindClosePointAlgorithm.kd_tree_scipy:
        has_partner, n_partner = _find_close_points_scipy(
            point_coordinates, tol, **kwargs
        )
    elif algorithm is FindClosePointAlgorithm.brute_force_cython:
        has_partner, n_partner = _find_close_points_brute_force_cython(
            point_coordinates, tol, **kwargs
        )
    elif algorithm is FindClosePointAlgorithm.boundary_volume_hierarchy_arborx:
        has_partner, n_partner = _find_close_points_arborx(
            point_coordinates, tol, **kwargs
        )
    else:
        raise TypeError(f"Got unexpected algorithm {algorithm}")

    return has_partner, n_partner



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines utility functions for meshes."""

from typing import Dict as _Dict
from typing import List as _List
from typing import Tuple as _Tuple

from beamme.core.conf import bme as _bme
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.node import Node as _Node


def get_coupled_nodes_to_master_map(
    mesh: _Mesh, *, assign_i_global: bool = False
) -> _Tuple[_Dict[_Node, _Node], _List[_Node]]:
    """Get a mapping of nodes in a mesh that should be "replaced" because they
    are coupled via a joint.

    In some finite element (FE) solvers, nodes coupled via joints are resolved
    by assigning a "master" node to represent the joint. This function identifies
    such nodes and creates a mapping where each coupled node is mapped to its
    master node.

    Args
    ----
    mesh:
        Input mesh
    assign_i_global:
        If this flag is set, the global indices are set in the node objects.

    Return
    ----
    replaced_node_to_master_map:
        A dictionary mapping each "replaced" node to its "master" node.
    unique_nodes:
        A list containing all unique nodes in the mesh, i.e., all nodes which
        are not coupled and the master nodes.
    """

    # Get a dictionary that maps the "replaced" nodes to the "master" ones
    replaced_node_to_master_map = {}
    for coupling in mesh.boundary_conditions[_bme.bc.point_coupling, _bme.geo.point]:
        if coupling.data is not _bme.coupling_dof.fix:
            raise ValueError(
                "This function is only implemented for rigid joints at the DOFs"
            )
        coupling_nodes = coupling.geometry_set.get_points()
        for node in coupling_nodes[1:]:
            replaced_node_to_master_map[node] = coupling_nodes[0]

    # Check that no "replaced" node is a "master" node
    master_nodes = set(replaced_node_to_master_map.values())
    for replaced_node in replaced_node_to_master_map.keys():
        if replaced_node in master_nodes:
            raise ValueError(
                "A replaced node is also a master nodes. This is not supported"
            )

    # Get all unique nodes
    unique_nodes = [
        node for node in mesh.nodes if node not in replaced_node_to_master_map
    ]

    # Optionally number the nodes
    if assign_i_global:
        for i_node, node in enumerate(unique_nodes):
            node.i_global = i_node
        for replaced_node, master_node in replaced_node_to_master_map.items():
            replaced_node.i_global = master_node.i_global

    # Return the mapping
    return replaced_node_to_master_map, unique_nodes



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file provides the mappings between BeamMe objects and 4C input
files."""

from typing import Any as _Any

from beamme.core.conf import bme as _bme
from beamme.core.element_volume import (
    VolumeHEX8 as _VolumeHEX8,
)
from beamme.core.element_volume import (
    VolumeHEX20 as _VolumeHEX20,
)
from beamme.core.element_volume import (
    VolumeHEX27 as _VolumeHEX27,
)
from beamme.core.element_volume import (
    VolumeTET4 as _VolumeTET4,
)
from beamme.core.element_volume import (
    VolumeTET10 as _VolumeTET10,
)
from beamme.core.element_volume import (
    VolumeWEDGE6 as _VolumeWEDGE6,
)
from beamme.core.nurbs_patch import NURBSSurface as _NURBSSurface
from beamme.core.nurbs_patch import NURBSVolume as _NURBSVolume
from beamme.four_c.element_volume import SolidRigidSphere as _SolidRigidSphere
from beamme.four_c.four_c_types import BeamType as _BeamType

INPUT_FILE_MAPPINGS: dict[str, _Any] = {}
INPUT_FILE_MAPPINGS["beam_types"] = {
    _BeamType.reissner: "BEAM3R",
    _BeamType.kirchhoff: "BEAM3K",
    _BeamType.euler_bernoulli: "BEAM3EB",
}
INPUT_FILE_MAPPINGS["boundary_conditions"] = {
    (_bme.bc.dirichlet, _bme.geo.point): "DESIGN POINT DIRICH CONDITIONS",
    (_bme.bc.dirichlet, _bme.geo.line): "DESIGN LINE DIRICH CONDITIONS",
    (_bme.bc.dirichlet, _bme.geo.surface): "DESIGN SURF DIRICH CONDITIONS",
    (_bme.bc.dirichlet, _bme.geo.volume): "DESIGN VOL DIRICH CONDITIONS",
    (_bme.bc.locsys, _bme.geo.point): "DESIGN POINT LOCSYS CONDITIONS",
    (_bme.bc.locsys, _bme.geo.line): "DESIGN LINE LOCSYS CONDITIONS",
    (_bme.bc.locsys, _bme.geo.surface): "DESIGN SURF LOCSYS CONDITIONS",
    (_bme.bc.locsys, _bme.geo.volume): "DESIGN VOL LOCSYS CONDITIONS",
    (_bme.bc.neumann, _bme.geo.point): "DESIGN POINT NEUMANN CONDITIONS",
    (_bme.bc.neumann, _bme.geo.line): "DESIGN LINE NEUMANN CONDITIONS",
    (_bme.bc.neumann, _bme.geo.surface): "DESIGN SURF NEUMANN CONDITIONS",
    (_bme.bc.neumann, _bme.geo.volume): "DESIGN VOL NEUMANN CONDITIONS",
    (
        _bme.bc.moment_euler_bernoulli,
        _bme.geo.point,
    ): "DESIGN POINT MOMENT EB CONDITIONS",
    (
        _bme.bc.beam_to_solid_volume_meshtying,
        _bme.geo.line,
    ): "BEAM INTERACTION/BEAM TO SOLID VOLUME MESHTYING LINE",
    (
        _bme.bc.beam_to_solid_volume_meshtying,
        _bme.geo.volume,
    ): "BEAM INTERACTION/BEAM TO SOLID VOLUME MESHTYING VOLUME",
    (
        _bme.bc.beam_to_solid_surface_meshtying,
        _bme.geo.line,
    ): "BEAM INTERACTION/BEAM TO SOLID SURFACE MESHTYING LINE",
    (
        _bme.bc.beam_to_solid_surface_meshtying,
        _bme.geo.surface,
    ): "BEAM INTERACTION/BEAM TO SOLID SURFACE MESHTYING SURFACE",
    (
        _bme.bc.beam_to_solid_surface_contact,
        _bme.geo.line,
    ): "BEAM INTERACTION/BEAM TO SOLID SURFACE CONTACT LINE",
    (
        _bme.bc.beam_to_solid_surface_contact,
        _bme.geo.surface,
    ): "BEAM INTERACTION/BEAM TO SOLID SURFACE CONTACT SURFACE",
    (_bme.bc.point_coupling, _bme.geo.point): "DESIGN POINT COUPLING CONDITIONS",
    (
        _bme.bc.beam_to_beam_contact,
        _bme.geo.line,
    ): "BEAM INTERACTION/BEAM TO BEAM CONTACT CONDITIONS",
    (
        _bme.bc.point_coupling_penalty,
        _bme.geo.point,
    ): "DESIGN POINT PENALTY COUPLING CONDITIONS",
    (
        _bme.bc.point_coupling_indirect,
        _bme.geo.line,
    ): "BEAM INTERACTION/BEAM TO BEAM POINT COUPLING CONDITIONS",
    (
        "DESIGN SURF MORTAR CONTACT CONDITIONS 3D",
        _bme.geo.surface,
    ): "DESIGN SURF MORTAR CONTACT CONDITIONS 3D",
}
INPUT_FILE_MAPPINGS["element_type_to_four_c_string"] = {
    _VolumeHEX8: "HEX8",
    _VolumeHEX20: "HEX20",
    _VolumeHEX27: "HEX27",
    _VolumeTET4: "TET4",
    _VolumeTET10: "TET10",
    _VolumeWEDGE6: "WEDGE6",
    _SolidRigidSphere: "POINT1",
}
INPUT_FILE_MAPPINGS["element_four_c_string_to_type"] = {
    value: key
    for key, value in INPUT_FILE_MAPPINGS["element_type_to_four_c_string"].items()
}
INPUT_FILE_MAPPINGS["geometry_sets_geometry_to_condition_name"] = {
    _bme.geo.point: "DNODE-NODE TOPOLOGY",
    _bme.geo.line: "DLINE-NODE TOPOLOGY",
    _bme.geo.surface: "DSURF-NODE TOPOLOGY",
    _bme.geo.volume: "DVOL-NODE TOPOLOGY",
}
INPUT_FILE_MAPPINGS["geometry_sets_condition_to_geometry_name"] = {
    value: key
    for key, value in INPUT_FILE_MAPPINGS[
        "geometry_sets_geometry_to_condition_name"
    ].items()
}
INPUT_FILE_MAPPINGS["geometry_sets_geometry_to_entry_name"] = {
    _bme.geo.point: "DNODE",
    _bme.geo.line: "DLINE",
    _bme.geo.surface: "DSURFACE",
    _bme.geo.volume: "DVOL",
}
INPUT_FILE_MAPPINGS["n_nodes_to_cell_type"] = {
    2: "LINE2",
    3: "LINE3",
    4: "LINE4",
    5: "LINE5",
}
INPUT_FILE_MAPPINGS["n_nodes_to_node_ordering"] = {
    2: [0, 1],
    3: [0, 2, 1],
    4: [0, 3, 1, 2],
    5: [0, 4, 1, 2, 3],
}
INPUT_FILE_MAPPINGS["nurbs_type_to_default_four_c_type"] = {
    _NURBSSurface: "WALLNURBS",
    _NURBSVolume: "SOLID",
}



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test vtk writer functionality of BeamMe."""

from unittest.mock import patch

import numpy as np
import vtk

from beamme.core.element_beam import Beam3
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.core.vtk_writer import VTKType, VTKWriter
from beamme.four_c.model_importer import import_four_c_model
from beamme.mesh_creation_functions.applications.beam_honeycomb import (
    create_beam_mesh_honeycomb,
)
from beamme.mesh_creation_functions.beam_arc import (
    create_beam_mesh_arc_segment_via_rotation,
)
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_integration_core_vtk_writer_write_vtk(
    assert_results_close, get_corresponding_reference_file_path, tmp_path
):
    """Test the output created by the VTK writer."""

    # Initialize writer.
    writer = VTKWriter()

    # Add poly line.
    indices = writer.add_points([[0, 0, -2], [1, 1, -2], [2, 2, -1]])
    writer.add_cell(vtk.vtkPolyLine, indices)

    # Add quadratic quad.
    cell_data = {}
    cell_data["cell_data_1"] = 3
    cell_data["cell_data_2"] = [66, 0, 1]
    point_data = {}
    point_data["point_data_1"] = [1, 2, 3, 4, 5, -2, -3, 0]
    point_data["point_data_2"] = [
        [0.25, 0, -0.25],
        [1, 0.25, 0],
        [2, 0, 0],
        [2.25, 1.25, 0.5],
        [2, 2.25, 0],
        [1, 2, 0.5],
        [0, 2.25, 0],
        [0, 1, 0.5],
    ]
    indices = writer.add_points(
        [
            [0.25, 0, -0.25],
            [1, 0.25, 0],
            [2, 0, 0],
            [2.25, 1.25, 0.5],
            [2, 2.25, 0],
            [1, 2, 0.5],
            [0, 2.25, 0],
            [0, 1, 0.5],
        ],
        point_data=point_data,
    )
    writer.add_cell(
        vtk.vtkQuadraticQuad, indices[[0, 2, 4, 6, 1, 3, 5, 7]], cell_data=cell_data
    )

    # Add tetrahedron.
    cell_data = {}
    cell_data["cell_data_2"] = [5, 0, 10]
    point_data = {}
    point_data["point_data_1"] = [1, 2, 3, 4]
    indices = writer.add_points(
        [[3, 3, 3], [4, 4, 3], [4, 3, 3], [4, 4, 4]], point_data=point_data
    )
    writer.add_cell(vtk.vtkTetra, indices[[0, 2, 1, 3]], cell_data=cell_data)

    # Before we can write the data to file we have to store the cell and
    # point data in the grid
    writer.complete_data()

    # Write to file.
    ref_file = get_corresponding_reference_file_path(extension="vtu")
    vtk_file = tmp_path / ref_file.name
    writer.write_vtk(vtk_file, binary=False)

    # Compare the vtk files.
    assert_results_close(ref_file, vtk_file)


def test_integration_core_vtk_writer_beam(
    assert_results_close,
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    tmp_path,
):
    """Create a sample mesh and check the VTK output."""

    # Create the mesh.
    mesh = Mesh()

    # Add content to the mesh.
    honeycomb_set = create_beam_mesh_honeycomb(
        mesh,
        Beam3,
        get_default_test_beam_material(),
        2.0,
        2,
        3,
        n_el=2,
    )
    mesh.add(honeycomb_set)

    # Write VTK output, with coupling sets."""
    ref_file = get_corresponding_reference_file_path(extension="vtu")
    vtk_file = tmp_path / "test_1_beam.vtu"
    mesh.write_vtk(
        output_name="test_1",
        output_directory=tmp_path,
        binary=False,
        coupling_sets=True,
    )
    assert_results_close(ref_file, vtk_file)

    # Write VTK output, without coupling sets."""
    ref_file = get_corresponding_reference_file_path(
        additional_identifier="no_coupling_beam", extension="vtu"
    )
    vtk_file = tmp_path / "test_2_beam.vtu"
    mesh.write_vtk(
        output_name="test_2",
        output_directory=tmp_path,
        binary=False,
        coupling_sets=False,
    )
    assert_results_close(ref_file, vtk_file)

    # Write VTK output, with coupling sets and additional points for visualization."""
    ref_file = get_corresponding_reference_file_path(
        additional_identifier="smooth_centerline_beam", extension="vtu"
    )
    vtk_file = tmp_path / "test_3_beam.vtu"
    mesh.write_vtk(
        output_name="test_3",
        coupling_sets=True,
        output_directory=tmp_path,
        binary=False,
        beam_centerline_visualization_segments=3,
    )
    assert_results_close(ref_file, vtk_file)


def test_integration_core_vtk_writer_solid(
    assert_results_close, get_corresponding_reference_file_path, tmp_path
):
    """Import a solid mesh and check the VTK output."""

    # Convert the solid mesh to beamme objects.
    _, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_tube"
        ),
        convert_input_to_mesh=True,
    )

    # Write VTK output.
    ref_file = get_corresponding_reference_file_path(extension="vtu")
    vtk_file = tmp_path / "test_solid.vtu"
    mesh.write_vtk(output_name="test", output_directory=tmp_path, binary=False)

    # Compare the vtk files.
    assert_results_close(ref_file, vtk_file)


def test_integration_core_vtk_writer_solid_elements(
    assert_results_close, get_corresponding_reference_file_path, tmp_path
):
    """Import a solid mesh with all solid types and check the VTK output."""

    # Convert the solid mesh to beamme objects.
    _, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            additional_identifier="import"
        ),
        convert_input_to_mesh=True,
    )

    # Write VTK output.
    ref_file = get_corresponding_reference_file_path(
        additional_identifier="solid", extension="vtu"
    )
    vtk_file = tmp_path / "test_solid.vtu"
    mesh.write_vtk(output_name="test", output_directory=tmp_path, binary=False)

    # Compare the vtk files.
    assert_results_close(ref_file, vtk_file)


def test_integration_core_vtk_writer_curve_cell_data(
    assert_results_close,
    get_default_test_beam_material,
    get_corresponding_reference_file_path,
    tmp_path,
):
    """Test that when creating a beam, cell data can be given.

    This test also checks, that the nan values in vtk can be explicitly
    given.
    """

    with (
        patch("beamme.core.vtk_writer.VTK_NAN_FLOAT", 69.69),
        patch("beamme.core.vtk_writer.VTK_NAN_INT", 69),
    ):
        # Create the mesh.
        mesh = Mesh()

        # Add content to the mesh.
        mat = get_default_test_beam_material()
        create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [2, 0, 0], n_el=2)
        create_beam_mesh_line(
            mesh,
            Beam3,
            mat,
            [0, 1, 0],
            [2, 1, 0],
            n_el=2,
            vtk_cell_data={"cell_data": (1, VTKType.int)},
        )
        create_beam_mesh_arc_segment_via_rotation(
            mesh,
            Beam3,
            mat,
            [0, 2, 0],
            Rotation([1, 0, 0], np.pi),
            1.5,
            np.pi / 2.0,
            n_el=2,
            vtk_cell_data={"cell_data": (2, VTKType.int), "other_data": 69},
        )

        # Write VTK output, with coupling sets."""
        ref_file = get_corresponding_reference_file_path(
            additional_identifier="beam", extension="vtu"
        )
        vtk_file = tmp_path / "test_beam.vtu"
        mesh.write_vtk(output_name="test", output_directory=tmp_path, binary=False)

        # Compare the vtk files.
        assert_results_close(ref_file, vtk_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the functionality of the core mesh utils."""

from beamme.core.element_beam import Beam2
from beamme.core.material import MaterialBeamBase
from beamme.core.mesh import Mesh
from beamme.core.mesh_utils import get_coupled_nodes_to_master_map
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_beamme_core_mesh_utils_get_coupled_nodes_to_master_map():
    """Test the get_coupled_nodes_to_master_map function."""

    beam_class = Beam2
    mat = MaterialBeamBase()
    mesh = Mesh()
    create_beam_mesh_line(mesh, beam_class, mat, [0, 0, 0], [1, 0, 0])
    create_beam_mesh_line(mesh, beam_class, mat, [1, 0, 0], [1, 1, 0])
    create_beam_mesh_line(mesh, beam_class, mat, [1, 1, 0], [2, 0, 0])
    create_beam_mesh_line(mesh, beam_class, mat, [1, 0, 0], [2, 0, 0])

    mesh.couple_nodes()

    replaced_node_to_master_map, unique_nodes = get_coupled_nodes_to_master_map(
        mesh, assign_i_global=True
    )

    assert len(unique_nodes) == 4
    assert unique_nodes == [mesh.nodes[i_node] for i_node in [0, 1, 3, 5]]

    assert len(replaced_node_to_master_map) == 4
    assert replaced_node_to_master_map[mesh.nodes[2]] == mesh.nodes[1]
    assert replaced_node_to_master_map[mesh.nodes[4]] == mesh.nodes[3]
    assert replaced_node_to_master_map[mesh.nodes[7]] == mesh.nodes[5]
    assert replaced_node_to_master_map[mesh.nodes[6]] == mesh.nodes[1]

    expected_i_global = [0, 1, 1, 2, 2, 3, 1, 3]
    for i_node, expected_index in enumerate(expected_i_global):
        assert mesh.nodes[i_node].i_global == expected_index



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module contains functions to load and parse existing 4C input files."""

from collections import defaultdict as _defaultdict
from pathlib import Path as _Path
from typing import Tuple as _Tuple

from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.boundary_condition import (
    BoundaryConditionBase as _BoundaryConditionBase,
)
from beamme.core.conf import bme as _bme
from beamme.core.coupling import Coupling as _Coupling
from beamme.core.geometry_set import GeometrySetNodes as _GeometrySetNodes
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.node import Node as _Node
from beamme.four_c.input_file import InputFile as _InputFile
from beamme.four_c.input_file_mappings import (
    INPUT_FILE_MAPPINGS as _INPUT_FILE_MAPPINGS,
)
from beamme.four_c.material import MaterialSolid as _MaterialSolid
from beamme.utils.environment import cubitpy_is_available as _cubitpy_is_available

if _cubitpy_is_available():
    from cubitpy.cubit_to_fourc_input import (
        get_input_file_with_mesh as _get_input_file_with_mesh,
    )


def import_cubitpy_model(
    cubit, convert_input_to_mesh: bool = False
) -> _Tuple[_InputFile, _Mesh]:
    """Convert a CubitPy instance to a BeamMe InputFile.

    Args:
        cubit (CubitPy): An instance of a cubit model.
        convert_input_to_mesh: If this is false, the cubit model will be
            converted to plain FourCIPP input data. If this is true, an input
            file with all the parameters will be returned and a mesh which
            contains the mesh information from cubit converted to BeamMe
            objects.

    Returns:
        A tuple with the input file and the mesh. If convert_input_to_mesh is
        False, the mesh will be empty. Note that the input sections which are
        converted to a BeamMe mesh are removed from the input file object.
    """

    input_file = _InputFile()
    input_file.add(_get_input_file_with_mesh(cubit).sections)

    if convert_input_to_mesh:
        return _extract_mesh_sections(input_file)
    else:
        return input_file, _Mesh()


def import_four_c_model(
    input_file_path: _Path, convert_input_to_mesh: bool = False
) -> _Tuple[_InputFile, _Mesh]:
    """Import an existing 4C input file and optionally convert it into a BeamMe
    mesh.

    Args:
        input_file_path: A file path to an existing 4C input file that will be
            imported.
        convert_input_to_mesh: If True, the input file will be converted to a
            BeamMe mesh.

    Returns:
        A tuple with the input file and the mesh. If convert_input_to_mesh is
        False, the mesh will be empty. Note that the input sections which are
        converted to a BeamMe mesh are removed from the input file object.
    """

    input_file = _InputFile().from_4C_yaml(input_file_path=input_file_path)

    if convert_input_to_mesh:
        return _extract_mesh_sections(input_file)
    else:
        return input_file, _Mesh()


def _extract_mesh_sections(input_file: _InputFile) -> _Tuple[_InputFile, _Mesh]:
    """Convert an InputFile into a native mesh by translating sections like
    materials, nodes, elements, geometry sets, and boundary conditions.

    Args:
        input_file: The input file containing 4C sections.
    Returns:
        A tuple (input_file, mesh). The input_file is modified in place to remove
        sections converted into BeamMe objects.
    """

    # function to pop sections from the input file
    _pop_section = lambda name: input_file.pop(name, [])

    # convert all sections to native objects and add to a new mesh
    mesh = _Mesh()

    # extract materials
    material_id_map_all = {}

    for mat in _pop_section("MATERIALS"):
        mat_id = mat.pop("MAT")
        if len(mat) != 1:
            raise ValueError(
                f"Could not convert the material data `{mat}` to a BeamMe material!"
            )
        mat_name, mat_data = list(mat.items())[0]
        material = _MaterialSolid(material_string=mat_name, data=mat_data)
        material_id_map_all[mat_id] = material

    nested_materials = set()
    for material in material_id_map_all.values():
        # Loop over each material and link nested materials. Also, mark nested materials
        # as they will not be added to the mesh.
        material_ids = material.data.get("MATIDS", [])
        for i_sub_material, material_id in enumerate(material_ids):
            try:
                material_ids[i_sub_material] = material_id_map_all[material_id]
            except KeyError as key_exception:
                raise KeyError(
                    f"Material ID {material_id} not in material_id_map_all (available "
                    f"IDs: {list(material_id_map_all.keys())})."
                ) from key_exception
            nested_materials.add(material_id)

    # Get a map of all non-nested materials. We assume that only those are used as
    # materials for elements. Also, add the non-nested materials to the mesh.
    material_id_map = {
        key: val
        for key, val in material_id_map_all.items()
        if key not in nested_materials
    }
    mesh.materials.extend(material_id_map.values())

    # extract nodes
    mesh.nodes = [_Node(node["COORD"]) for node in _pop_section("NODE COORDS")]

    # extract elements
    for input_element in _pop_section("STRUCTURE ELEMENTS"):
        if (
            input_element["cell"]["type"]
            not in _INPUT_FILE_MAPPINGS["element_four_c_string_to_type"]
        ):
            raise TypeError(
                f"Could not create a BeamMe element for `{input_element['data']['type']}` `{input_element['cell']['type']}`!"
            )
        nodes = [mesh.nodes[i - 1] for i in input_element["cell"]["connectivity"]]
        element_class = _INPUT_FILE_MAPPINGS["element_four_c_string_to_type"][
            input_element["cell"]["type"]
        ]
        element = element_class(nodes=nodes, data=input_element["data"])
        if "MAT" in element.data:
            element.data["MAT"] = material_id_map[element.data["MAT"]]
        mesh.elements.append(element)

    # extract geometry sets
    geometry_sets_in_sections: dict[str, dict[int, _GeometrySetNodes]] = _defaultdict(
        dict
    )

    for section_name in input_file.sections:
        if not section_name.endswith("TOPOLOGY"):
            continue

        items = _pop_section(section_name)
        if not items:
            continue

        # Find geometry type for this section
        try:
            geometry_type = _INPUT_FILE_MAPPINGS[
                "geometry_sets_condition_to_geometry_name"
            ][section_name]
        except ValueError as e:
            raise ValueError(f"Unknown geometry section: {section_name}") from e

        # Extract geometry set indices
        geom_dict: dict[int, list[int]] = _defaultdict(list)
        for entry in items:
            geom_dict[entry["d_id"]].append(entry["node_id"] - 1)

        geometry_sets_in_sections[geometry_type] = {
            gid: _GeometrySetNodes(geometry_type, nodes=[mesh.nodes[i] for i in ids])
            for gid, ids in geom_dict.items()
        }

        mesh.geometry_sets[geometry_type] = list(
            geometry_sets_in_sections[geometry_type].values()
        )

    # extract boundary conditions
    _standard_bc_types = (
        _bme.bc.dirichlet,
        _bme.bc.neumann,
        _bme.bc.locsys,
        _bme.bc.beam_to_solid_surface_meshtying,
        _bme.bc.beam_to_solid_surface_contact,
        _bme.bc.beam_to_solid_volume_meshtying,
    )

    for (bc_key, geometry_type), section_name in _INPUT_FILE_MAPPINGS[
        "boundary_conditions"
    ].items():
        for bc_data in _pop_section(section_name):
            geometry_set = geometry_sets_in_sections[geometry_type][bc_data.pop("E")]

            bc_obj: _BoundaryConditionBase

            if bc_key in _standard_bc_types or isinstance(bc_key, str):
                bc_obj = _BoundaryCondition(geometry_set, bc_data, bc_type=bc_key)
            elif bc_key is _bme.bc.point_coupling:
                bc_obj = _Coupling(
                    geometry_set, bc_key, bc_data, check_overlapping_nodes=False
                )
            else:
                raise ValueError(f"Unexpected boundary condition: {bc_key}")

            mesh.boundary_conditions.append((bc_key, geometry_type), bc_obj)

    return input_file, mesh



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Create a beam filament from a curve represented with splinepy."""

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.mesh_creation_functions.beam_parametric_curve import (
    create_beam_mesh_parametric_curve as _create_beam_mesh_parametric_curve,
)


def get_curve_function_and_jacobian_for_integration(curve, tol: float | None = None):
    """Return function objects for evaluating the curve and the derivative.
    These functions are used in the curve integration. It can happen that the
    integration algorithm has to evaluate the curve outside of the defined
    domain. This usually leads to errors in common spline/NURBS packages.
    Therefore, we check for this evaluation outside of the parameter domain
    here and perform a linear extrapolation.

    Args
    ----
    curve: splinepy object
        Curve that is used to describe the beam centerline.
    tol: float
        Tolerance for checking if point is close to the start or end of the
        interval. If None is given, use the default tolerance from bme.

    Return
    ----
    (function, jacobian, curve_start, curve_end):
        function:
            Function for evaluating a position on the curve
        jacobian:
            Function for evaluating the tangent along the curve
        curve_start:
            Parameter coordinate for the start for the curve
        curve_end:
            Parameter coordinate for the end for the curve
    """

    if tol is None:
        tol = _bme.eps_pos

    curve_start = curve.parametric_bounds[0][0]
    curve_end = curve.parametric_bounds[1][0]

    def eval_r(t):
        """Evaluate the position along the curve."""
        return curve.evaluate([[t]])[0]

    def eval_rp(t):
        """Evaluate the derivative along the curve."""
        return curve.derivative([[t]], orders=[1])[0]

    def function(t):
        """Convert the curve to a function that can be used for beam
        generation."""

        if curve_start <= t <= curve_end:
            return eval_r(t)
        elif t < curve_start and _np.abs(t - curve_start) < tol:
            diff = t - curve_start
            return eval_r(curve_start) + diff * eval_rp(curve_start)
        elif t > curve_end and _np.abs(t - curve_end) < tol:
            diff = t - curve_end
            return eval_r(curve_end) + diff * eval_rp(curve_end)
        raise ValueError(
            "Can not evaluate the curve function outside of the interval (plus tolerances).\n"
            f"Abs diff start: {_np.abs(curve_start - t)}\nAbs diff end: {_np.abs(t - curve_end)}"
        )

    def jacobian(t):
        """Convert the curve to a Jacobian function that can be used for
        integration along the curve.

        There is no tolerance here, since the integration algorithms
        sometimes evaluates the derivative far outside the interval.
        """

        if curve_start <= t <= curve_end:
            return eval_rp(t)
        elif t < curve_start:
            return eval_rp(curve_start)
        elif curve_end < t:
            return eval_rp(curve_end)
        raise ValueError("Should not happen")

    return function, jacobian, curve_start, curve_end


def create_beam_mesh_from_splinepy(
    mesh, beam_class, material, curve, *, tol=None, **kwargs
):
    """Generate a beam from a splinepy curve.

    Args
    ----
    mesh: Mesh
        Mesh that the curve will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this line.
    curve: splinepy object
        Curve that is used to describe the beam centerline.
    tol: float
        Tolerance for checking if point is close to the start or end of the
        interval. If None is given, use the default tolerance from bme.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return:
        Return value from create_beam_mesh_function
    """

    (
        function,
        jacobian,
        curve_start,
        curve_end,
    ) = get_curve_function_and_jacobian_for_integration(curve, tol=tol)

    # Create the beams
    return _create_beam_mesh_parametric_curve(
        mesh,
        beam_class,
        material,
        function,
        [curve_start, curve_end],
        function_derivative=jacobian,
        **kwargs,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Generic function used to create NURBS meshes."""

import itertools as _itertools
from typing import Type as _Type

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySetNodes as _GeometrySetNodes
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.node import ControlPoint as _ControlPoint
from beamme.core.nurbs_patch import NURBSSurface as _NURBSSurface
from beamme.core.nurbs_patch import NURBSVolume as _NURBSVolume


def add_splinepy_nurbs_to_mesh(
    mesh: _Mesh,
    splinepy_obj,
    *,
    material=None,
    data: dict | None = None,
) -> _GeometryName:
    """Add a splinepy NURBS to the mesh.

    Args:
        mesh: Mesh that the created NURBS geometry will be added to.
        splinepy_obj (splinepy object): NURBS geometry created using splinepy.
        material (Material): Material for this geometry.
        data: General element data, e.g., material, formulation, ...

    Returns:
        GeometryName:
            Set with the control points that form the topology of the mesh.

            For a surface, the following information is stored:
                Vertices: 'vertex_u_min_v_min', 'vertex_u_max_v_min', 'vertex_u_min_v_max', 'vertex_u_max_v_max'
                Edges: 'line_v_min', 'line_u_max', 'line_v_max', 'line_u_min'
                Surface: 'surf'

            For a volume, the following information is stored:
                Vertices: 'vertex_u_min_v_min_w_min', 'vertex_u_max_v_min_w_min', 'vertex_u_min_v_max_w_min', 'vertex_u_max_v_max_w_min',
                        'vertex_u_min_v_min_w_max', 'vertex_u_max_v_min_w_max', 'vertex_u_min_v_max_w_max', 'vertex_u_max_v_max_w_max'
                Edges: 'line_v_min_w_min', 'line_u_max_w_min', 'line_v_max_w_min', 'line_u_min_w_min',
                    'line_u_min_v_min', 'line_u_max_v_min', 'line_u_min_v_max', 'line_u_max_v_max'
                    'line_v_min_w_max', 'line_u_max_w_max', 'line_v_max_w_max', 'line_u_min_w_max'
                Surfaces: 'surf_w_min', 'surf_w_max', 'surf_v_min', 'surf_v_max', 'surf_v_max', 'surf_u_min'
                Volume: 'vol'
    """

    # Make sure that the control points are 3D
    nurbs_cp_dim = splinepy_obj.control_points.shape[1]
    if not nurbs_cp_dim == 3:
        raise ValueError(f"Invalid control point dimension: {nurbs_cp_dim}")

    # Make sure the material is in the mesh
    mesh.add_material(material)

    # Fill control points
    control_points = [
        _ControlPoint(coord, weight[0])
        for coord, weight in zip(
            _np.asarray(splinepy_obj.control_points), _np.asarray(splinepy_obj.weights)
        )
    ]

    # Fill element
    manifold_dim = len(splinepy_obj.knot_vectors)
    nurbs_object: _Type[_NURBSSurface] | _Type[_NURBSVolume]
    if manifold_dim == 2:
        nurbs_object = _NURBSSurface
    elif manifold_dim == 3:
        nurbs_object = _NURBSVolume
    else:
        raise NotImplementedError(
            "Error, not implemented for a NURBS {}!".format(type(splinepy_obj))
        )

    element = nurbs_object(
        [_np.asarray(knot_vector) for knot_vector in splinepy_obj.knot_vectors],
        _np.asarray(splinepy_obj.degrees),
        nodes=control_points,
        material=material,
        data=data,
    )

    # Add element and control points to the mesh
    mesh.elements.append(element)
    mesh.nodes.extend(control_points)

    # Create geometry sets that will be returned
    return_set = create_geometry_sets(element)

    return return_set


def add_geomdl_nurbs_to_mesh(
    mesh: _Mesh,
    geomdl_obj,
    *,
    material=None,
    data: dict | None = None,
) -> _GeometryName:
    """Add a geomdl NURBS to the mesh.

    Args:
        mesh: Mesh that the created NURBS geometry will be added to.
        geomdl_obj (geomdl object): NURBS geometry created using geomdl.
        material (Material): Material for this geometry.
        data: General element data, e.g., material, formulation, ...

    Returns:
        GeometryName:
            Set with the control points that form the topology of the mesh.

            For a surface, the following information is stored:
                Vertices: 'vertex_u_min_v_min', 'vertex_u_max_v_min', 'vertex_u_min_v_max', 'vertex_u_max_v_max'
                Edges: 'line_v_min', 'line_u_max', 'line_v_max', 'line_u_min'
                Surface: 'surf'

            For a volume, the following information is stored:
                Vertices: 'vertex_u_min_v_min_w_min', 'vertex_u_max_v_min_w_min', 'vertex_u_min_v_max_w_min', 'vertex_u_max_v_max_w_min',
                        'vertex_u_min_v_min_w_max', 'vertex_u_max_v_min_w_max', 'vertex_u_min_v_max_w_max', 'vertex_u_max_v_max_w_max'
                Edges: 'line_v_min_w_min', 'line_u_max_w_min', 'line_v_max_w_min', 'line_u_min_w_min',
                    'line_u_min_v_min', 'line_u_max_v_min', 'line_u_min_v_max', 'line_u_max_v_max'
                    'line_v_min_w_max', 'line_u_max_w_max', 'line_v_max_w_max', 'line_u_min_w_max'
                Surfaces: 'surf_w_min', 'surf_w_max', 'surf_v_min', 'surf_v_max', 'surf_v_max', 'surf_u_min'
                Volume: 'vol'
    """

    # Make sure the material is in the mesh
    mesh.add_material(material)

    # Fill control points
    control_points = []
    nurbs_dimension = len(geomdl_obj.knotvector)
    if nurbs_dimension == 2:
        control_points = create_control_points_surface(geomdl_obj)
    elif nurbs_dimension == 3:
        control_points = create_control_points_volume(geomdl_obj)
    else:
        raise NotImplementedError(
            "Error, not implemented for NURBS with dimension {}!".format(
                nurbs_dimension
            )
        )

    # Fill element
    manifold_dim = len(geomdl_obj.knotvector)
    nurbs_object: _Type[_NURBSSurface] | _Type[_NURBSVolume]
    if manifold_dim == 2:
        nurbs_object = _NURBSSurface
    elif manifold_dim == 3:
        nurbs_object = _NURBSVolume
    else:
        raise NotImplementedError(
            "Error, not implemented for a NURBS {}!".format(type(geomdl_obj))
        )

    element = nurbs_object(
        geomdl_obj.knotvector,
        geomdl_obj.degree,
        nodes=control_points,
        material=material,
        data=data,
    )

    # Add element and control points to the mesh
    mesh.elements.append(element)
    mesh.nodes.extend(control_points)

    # Create geometry sets that will be returned
    return_set = create_geometry_sets(element)

    return return_set


def create_control_points_surface(geomdl_obj):
    """Creates a list with the ControlPoint objects of a surface created with
    geomdl."""
    control_points = []
    for dir_v in range(geomdl_obj.ctrlpts_size_v):
        for dir_u in range(geomdl_obj.ctrlpts_size_u):
            weight = geomdl_obj.ctrlpts2d[dir_u][dir_v][3]

            # As the control points are scaled with their weight, divide them to get
            # their coordinates
            coord = [
                geomdl_obj.ctrlpts2d[dir_u][dir_v][0] / weight,
                geomdl_obj.ctrlpts2d[dir_u][dir_v][1] / weight,
                geomdl_obj.ctrlpts2d[dir_u][dir_v][2] / weight,
            ]

            control_points.append(_ControlPoint(coord, weight))

    return control_points


def create_control_points_volume(geomdl_obj):
    """Creates a list with the ControlPoint objects of a volume created with
    geomdl."""
    control_points = []
    for dir_w in range(geomdl_obj.ctrlpts_size_w):
        for dir_v in range(geomdl_obj.ctrlpts_size_v):
            for dir_u in range(geomdl_obj.ctrlpts_size_u):
                # Obtain the id of the control point
                cp_id = (
                    dir_v
                    + geomdl_obj.ctrlpts_size_v * dir_u
                    + geomdl_obj.ctrlpts_size_u * geomdl_obj.ctrlpts_size_v * dir_w
                )

                weight = geomdl_obj.ctrlptsw[cp_id][3]

                # As the control points are scaled with their weight, divide them to get
                # their coordinates
                coord = [
                    geomdl_obj.ctrlptsw[cp_id][0] / weight,
                    geomdl_obj.ctrlptsw[cp_id][1] / weight,
                    geomdl_obj.ctrlptsw[cp_id][2] / weight,
                ]

                control_points.append(_ControlPoint(coord, weight))

    return control_points


def create_geometry_sets(element: _NURBSSurface | _NURBSVolume) -> _GeometryName:
    """Create the geometry sets for NURBS patches of all dimensions.

    Args:
        element: The NURBS patch for which the geometry sets should be created.

    Returns:
        The geometry set container for the given NURBS patch.
    """

    # Create return set
    return_set = _GeometryName()

    # Get general data needed for the set creation
    num_cps_uvw = element.get_number_of_control_points_per_dir()
    nurbs_dimension = len(element.knot_vectors)
    n_cp = _np.prod(num_cps_uvw)
    axes = ["u", "v", "w"][:nurbs_dimension]
    name_map = {0: "min", -1: "max", 1: "min_next", -2: "max_next"}

    # This is a tensor array that contains the CP indices
    cp_indices_dim = _np.arange(n_cp, dtype=int).reshape(*num_cps_uvw[::-1]).transpose()

    if nurbs_dimension >= 0:
        # Add point sets
        directions = [0, -1]
        for corner in _itertools.product(*([directions] * nurbs_dimension)):
            name = "vertex_" + "_".join(
                f"{axis}_{name_map[coord]}" for axis, coord in zip(axes, corner)
            )
            index = cp_indices_dim[corner]
            return_set[name] = _GeometrySetNodes(
                _bme.geo.point, nodes=element.nodes[index]
            )

    if nurbs_dimension > 0:
        # Add edge sets

        if nurbs_dimension == 2:
            directions = [0, 1, -2, -1]
        elif nurbs_dimension == 3:
            directions = [0, -1]
        else:
            raise ValueError("NURBS dimension 1 not implemented")

        # Iterate over each axis (the axis that varies — the "edge" direction)
        for edge_axis in range(nurbs_dimension):
            # The other axes will be fixed
            fixed_axes = [i for i in range(nurbs_dimension) if i != edge_axis]
            for fixed_dir in _itertools.product(*([directions] * len(fixed_axes))):
                # Build slicing tuple for indexing cp_indices_dim
                slicer: list[slice | int] = [slice(None)] * nurbs_dimension
                name_parts = []
                for axis_idx, dir_val in zip(fixed_axes, fixed_dir):
                    slicer[axis_idx] = dir_val
                    name_parts.append(f"{axes[axis_idx]}_{name_map[dir_val]}")
                name = "line_" + "_".join(name_parts)

                # Get node indices along the edge
                edge_indices = cp_indices_dim[tuple(slicer)].flatten()
                return_set[name] = _GeometrySetNodes(
                    _bme.geo.line, nodes=[element.nodes[i] for i in edge_indices]
                )

    if nurbs_dimension == 2:
        # Add surface sets for surface NURBS
        return_set["surf"] = _GeometrySetNodes(_bme.geo.surface, element.nodes)

    if nurbs_dimension == 3:
        # Add surface sets for volume NURBS
        for fixed_axis in range(nurbs_dimension):
            for dir_val in directions:
                # Build slice and name
                slicer = [slice(None)] * nurbs_dimension
                slicer[fixed_axis] = dir_val
                surface_name = f"surf_{axes[fixed_axis]}_{name_map[dir_val]}"

                surface_indices = cp_indices_dim[tuple(slicer)].flatten()
                return_set[surface_name] = _GeometrySetNodes(
                    _bme.geo.surface,
                    nodes=[element.nodes[i] for i in surface_indices],
                )

        # Add volume sets
        return_set["vol"] = _GeometrySetNodes(_bme.geo.volume, element.nodes)

    return return_set



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file provides functions to create Abaqus beam element classes."""

from beamme.core.material import MaterialBeamBase as _MaterialBeamBase


class AbaqusBeamMaterial(_MaterialBeamBase):
    """A class representing an Abaqus beam material."""

    def __init__(self, name: str):
        """Initialize the material. For now it is only supported to state the
        name of the resulting element set here. The material and cross-section
        lines in the input file have to be defined manually.

        Args
        ----
        name: str
            Name of the material, this will be the name of the resulting element set
        """
        super().__init__(data=name)

    def dump_to_list(self):
        """Return a list with the (single) item representing this material."""
        return [self.data]



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Testing framework infrastructure for result comparison."""

import json
import shutil
import subprocess
from pathlib import Path
from typing import Any, Callable

import numpy as np
import pytest
import vtk
import xmltodict
from fourcipp.utils.dict_utils import compare_nested_dicts_or_lists
from vistools.vtk.compare_grids import compare_grids

from beamme.core.mesh import Mesh
from beamme.four_c.input_file import InputFile

# GLOBAL DEFAULT TEST TOLERANCES
RELATIVE_TOLERANCE = 1e-13
ABSOLUTE_TOLERANCE = 1e-13


@pytest.fixture(scope="function")
def assert_results_close(tmp_path, current_test_name) -> Callable:
    """Return function to compare either string or files.

    Necessary to enable the function call through pytest fixtures.

    Args:
        tmp_path: Temporary path to write file if assertion fails.
        current_test_name: Name of the current test to create file
            if assertion fails.

    Returns:
        Function to compare results.
    """

    def _assert_results_close(
        reference: (
            Path | str | int | float | dict | list | np.ndarray | InputFile | Mesh
        ),
        result: Path | str | int | float | dict | list | np.ndarray | InputFile | Mesh,
        rtol: float = RELATIVE_TOLERANCE,
        atol: float = ABSOLUTE_TOLERANCE,
    ) -> None:
        """Comparison between reference and result with relative or absolute
        tolerance.

        If the comparison fails, an assertion is raised.

        Args:
            reference: The reference data.
            result: The result data.
            rtol: The relative tolerance.
            atol: The absolute tolerance.
        """

        # special case to compare vtk files
        if (
            isinstance(reference, Path)
            and reference.suffix in {".vtk", ".vtu"}
            and isinstance(result, Path)
            and result.suffix in {".vtk", ".vtu"}
        ):
            compare_vtk_files(reference, result, rtol, atol)
            return

        # convert all other types into dicts/lists
        converted_reference = convert_to_primitive_type(reference)
        converted_result = convert_to_primitive_type(result)

        try:
            compare_nested_dicts_or_lists(
                converted_reference,
                converted_result,
                rtol=rtol,
                atol=atol,
                allow_int_vs_float_comparison=True,
                custom_compare=lambda obj, ref_obj: custom_fourcipp_comparison(
                    obj, ref_obj, rtol=rtol, atol=atol
                ),
            )
        except AssertionError as error:
            handle_failed_assertion(tmp_path, current_test_name, reference, result)
            raise error

    return _assert_results_close


def compare_vtk_files(reference: Path, result: Path, rtol: float, atol: float) -> None:
    """Compare two VTK files for equality within a given tolerance.

    Args:
        reference: The path to the reference VTK file.
        result: The path to the result VTK file to be compared.
        rtol: The relative tolerance.
        atol: The absolute tolerance.
    """

    def get_vtk(path: Path) -> vtk.vtkDataObject:
        """Return vtk data object for given vtk file.

        Args:
            path: Path to .vtu/.vtk file.

        Returns:
            VTK data object.
        """

        reader = vtk.vtkXMLGenericDataObjectReader()
        reader.SetFileName(path)
        reader.Update()
        return reader.GetOutput()

    compare = compare_grids(
        get_vtk(reference), get_vtk(result), output=True, rtol=rtol, atol=atol
    )

    if not compare[0]:
        raise AssertionError("\n".join(compare[1]))


def convert_to_primitive_type(
    obj: str | int | float | dict | list | np.ndarray | Path | Mesh | InputFile,
) -> int | float | dict | list | np.ndarray:
    """Convert the given object to a primitive type (dict, list, numpy array).

    Args:
        obj: The object to convert.

    Returns:
        The raw data (either a dictionary, list, numpy array).
    """

    if isinstance(obj, (int, float, dict, list, np.ndarray)):
        return obj

    if isinstance(obj, Path):
        if obj.suffix == ".xml" or obj.suffix == ".pvd":
            return xmltodict.parse(obj.read_text(encoding="utf-8"))

        elif obj.suffix == ".json":
            return json.loads(obj.read_text(encoding="utf-8"))

        elif obj.name.endswith(".4C.yaml"):
            # return sections in next step
            obj = InputFile().from_4C_yaml(input_file_path=obj)

        elif obj.suffix == ".inp":
            # Abaqus input files are read as strings
            # split into list of fragments in next step
            obj = obj.read_text(encoding="utf-8")

    if isinstance(obj, Mesh):
        # Internally convert Mesh to InputFile to allow for simple comparison via dictionary
        # TODO this should be improved in the future to not fall back to use the 4C specific InputFile
        input_file = InputFile()
        input_file.add(obj)
        # return sections in next step
        obj = input_file

    if isinstance(obj, InputFile):
        return obj.sections

    if isinstance(obj, str):
        # Comparison for string based Abaqus input files
        # Split the string into individual fragments which can then be compared with tolerance

        def str_to_float(string: str) -> str | float:
            """Convert string to float if possible, otherwise return the
            string.

            Args:
                string: The string to convert.
            Returns:
                The converted string or float.
            """
            try:
                return float(string)
            except ValueError:
                return string

        return [
            str_to_float(fragment)
            for line in obj.splitlines()
            for fragment in line.split(",")
        ]

    raise TypeError(f"The comparison for {type(obj)} is not yet implemented!")


def custom_fourcipp_comparison(
    obj: Any, reference_obj: Any, rtol: float, atol: float
) -> bool | None:
    """Custom comparison function for the FourCIPP
    compare_nested_dicts_or_lists function.

    Comparison between two objects, either lists or numpy arrays.

    Args:
        obj: The object to compare.
        reference_obj: The reference object to compare against.

    Returns:
        True if the objects are equal, otherwise raises an AssertionError.
        If no comparison took place, None is returned.
    """

    if isinstance(obj, (np.ndarray, np.generic)) or isinstance(
        reference_obj, (np.ndarray, np.generic)
    ):
        if not np.allclose(obj, reference_obj, rtol=rtol, atol=atol):
            raise AssertionError(
                f"Custom BeamMe comparison failed!\n\nThe objects are not equal:\n\nobj: {obj}\n\nreference_obj: {reference_obj}"
            )
        return True

    return None


def handle_failed_assertion(
    tmp_path: Path,
    current_test_name: str,
    reference: Path | str | dict | list | np.ndarray | InputFile | Mesh,
    result: Path | str | dict | list | np.ndarray | InputFile | Mesh,
) -> None:
    """Handle failed assertions by opening a diff tool.

    For failed assertions the new result file is written to the temporary
    pytest directory and the VSCode diff tool is opened if available.

    Args:
        tmp_path: Temporary pytest directory.
        current_test_name: Name of the current test.
        reference: The reference data.
        result: The result data.
    """

    # if reference is not a file or if result is not a Mesh or InputFile we do not open the diff
    if not isinstance(reference, Path) or not isinstance(result, (Mesh, InputFile)):
        return

    # save result string to file
    result_path = tmp_path / (
        current_test_name + "_result" + "".join(reference.suffixes)
    )

    if isinstance(result, Mesh):
        input_file = InputFile()
        input_file.add(result)
        result = input_file

    result.dump(
        result_path,
        add_header_default=False,
        add_header_information=False,
        add_footer_application_script=False,
        validate=False,
    )

    print(f"Result string saved to: '{result_path}'.")

    # open VSCode diff tool if available
    if shutil.which("code") is not None:
        child = subprocess.Popen(
            ["code", "--diff", result_path, reference],
            stderr=subprocess.PIPE,
        )
        child.communicate()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines the classes that are used to create an input file for
4C."""

from __future__ import annotations as _annotations

import os as _os
from datetime import datetime as _datetime
from pathlib import Path as _Path
from typing import Any as _Any
from typing import Callable as _Callable
from typing import List as _List

from fourcipp.fourc_input import FourCInput as _FourCInput
from fourcipp.fourc_input import sort_by_section_names as _sort_by_section_names
from fourcipp.utils.not_set import NOT_SET as _NOT_SET

from beamme.core.conf import INPUT_FILE_HEADER as _INPUT_FILE_HEADER
from beamme.core.conf import bme as _bme
from beamme.core.function import Function as _Function
from beamme.core.material import Material as _Material
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.node import Node as _Node
from beamme.core.nurbs_patch import NURBSPatch as _NURBSPatch
from beamme.four_c.input_file_dump_item import dump_item_to_list as _dump_item_to_list
from beamme.four_c.input_file_dump_item import (
    dump_item_to_section as _dump_item_to_section,
)
from beamme.four_c.input_file_mappings import (
    INPUT_FILE_MAPPINGS as _INPUT_FILE_MAPPINGS,
)
from beamme.four_c.material import (
    get_all_contained_materials as _get_all_contained_materials,
)
from beamme.utils.environment import cubitpy_is_available as _cubitpy_is_available
from beamme.utils.environment import get_application_path as _get_application_path
from beamme.utils.environment import get_git_data as _get_git_data

if _cubitpy_is_available():
    import cubitpy as _cubitpy


class InputFile:
    """An item that represents a complete 4C input file."""

    def __init__(self):
        """Initialize the input file."""

        self.fourc_input = _FourCInput()

        # Contents of NOX xml file.
        self.nox_xml_contents = ""

        # Register converters to directly convert non-primitive types
        # to native Python types via the FourCIPP type converter.
        self.fourc_input.type_converter.register_numpy_types()
        self.fourc_input.type_converter.register_type(
            (_Function, _Material, _Node), lambda converter, obj: obj.i_global + 1
        )

    def __contains__(self, key: str) -> bool:
        """Contains function.

        Allows to use the `in` operator.

        Args:
            key: Section name to check if it is set

        Returns:
            True if section is set
        """

        return key in self.fourc_input

    def __setitem__(self, key: str, value: _Any) -> None:
        """Set section.

        Args:
            key: Section name
            value: Section entry
        """

        self.fourc_input[key] = value

    def __getitem__(self, key: str) -> _Any:
        """Get section of input file.

        Allows to use the indexing operator.

        Args:
            key: Section name to get

        Returns:
            The section content
        """

        return self.fourc_input[key]

    @classmethod
    def from_4C_yaml(
        cls, input_file_path: str | _Path, header_only: bool = False
    ) -> InputFile:
        """Load 4C yaml file.

        Args:
            input_file_path: Path to yaml file
            header_only: Only extract header, i.e., all sections except the legacy ones

        Returns:
            Initialised object
        """

        obj = cls()
        obj.fourc_input = _FourCInput.from_4C_yaml(input_file_path, header_only)
        return obj

    @property
    def sections(self) -> dict:
        """All the set sections.

        Returns:
            dict: Set sections
        """

        return self.fourc_input.sections

    def pop(self, key: str, default_value: _Any = _NOT_SET) -> _Any:
        """Pop section of input file.

        Args:
            key: Section name to pop

        Returns:
            The section content
        """

        return self.fourc_input.pop(key, default_value)

    def add(self, object_to_add, **kwargs):
        """Add a mesh or a dictionary to the input file.

        Args:
            object: The object to be added. This can be a mesh or a dictionary.
            **kwargs: Additional arguments to be passed to the add method.
        """

        if isinstance(object_to_add, _Mesh):
            self.add_mesh_to_input_file(mesh=object_to_add, **kwargs)

        else:
            self.fourc_input.combine_sections(object_to_add)

    def dump(
        self,
        input_file_path: str | _Path,
        *,
        nox_xml_file: str | None = None,
        add_header_default: bool = True,
        add_header_information: bool = True,
        add_footer_application_script: bool = True,
        validate=True,
        validate_sections_only: bool = False,
        sort_function: _Callable[[dict], dict] | None = _sort_by_section_names,
        fourcipp_yaml_style: bool = True,
    ):
        """Write the input file to disk.

        Args:
            input_file_path:
                Path to the input file that should be created.
            nox_xml_file:
                If this is a string, the NOX xml file will be created with this
                name. If this is None, the NOX xml file will be created with the
                name of the input file with the extension ".nox.xml".
            add_header_default:
                Prepend the default header comment to the input file.
            add_header_information:
                If the information header should be exported to the input file
                Contains creation date, git details of BeamMe, CubitPy and
                original application which created the input file if available.
            add_footer_application_script:
                Append the application script which creates the input files as a
                comment at the end of the input file.
            validate:
                Validate if the created input file is compatible with 4C with FourCIPP.
            validate_sections_only:
                Validate each section independently. Required sections are no longer
                required, but the sections must be valid.
            sort_function:
                A function which sorts the sections of the input file.
            fourcipp_yaml_style:
                If True, the input file is written in the fourcipp yaml style.
        """

        # Make sure the given input file is a Path instance.
        input_file_path = _Path(input_file_path)

        if self.nox_xml_contents:
            if nox_xml_file is None:
                nox_xml_file = input_file_path.name.split(".")[0] + ".nox.xml"

            self["STRUCT NOX/Status Test"] = {"XML File": nox_xml_file}

            # Write the xml file to the disc.
            with open(input_file_path.parent / nox_xml_file, "w") as xml_file:
                xml_file.write(self.nox_xml_contents)

        # Add information header to the input file
        if add_header_information:
            self.add({"TITLE": self._get_header()})

        self.fourc_input.dump(
            input_file_path=input_file_path,
            validate=validate,
            validate_sections_only=validate_sections_only,
            convert_to_native_types=False,  # conversion already happens during add()
            sort_function=sort_function,
            use_fourcipp_yaml_style=fourcipp_yaml_style,
        )

        if add_header_default or add_footer_application_script:
            with open(input_file_path, "r") as input_file:
                lines = input_file.readlines()

                if add_header_default:
                    lines = ["# " + line + "\n" for line in _INPUT_FILE_HEADER] + lines

                if add_footer_application_script:
                    application_path = _get_application_path()
                    if application_path is not None:
                        lines += self._get_application_script(application_path)

                with open(input_file_path, "w") as input_file:
                    input_file.writelines(lines)

    def add_mesh_to_input_file(self, mesh: _Mesh) -> None:
        """Add a mesh to the input file.

        Args:
            mesh: The mesh to be added to the input file.
        """

        if _bme.check_overlapping_elements:
            mesh.check_overlapping_elements()

        # Compute geometry-set starting indices
        start_indices_geometry_set = {
            geometry_type: max(
                (entry["d_id"] for entry in self.sections.get(section_name, [])),
                default=0,
            )
            for geometry_type, section_name in _INPUT_FILE_MAPPINGS[
                "geometry_sets_geometry_to_condition_name"
            ].items()
        }

        # Determine global start indices
        start_index_nodes = len(self.sections.get("NODE COORDS", []))

        start_index_elements = sum(
            len(self.sections.get(section, []))
            for section in ("FLUID ELEMENTS", "STRUCTURE ELEMENTS")
        )

        start_index_functions = max(
            (
                int(section.split("FUNCT")[-1])
                for section in self.sections
                if section.startswith("FUNCT")
            ),
            default=0,
        )

        start_index_materials = max(
            (material["MAT"] for material in self.sections.get("MATERIALS", [])),
            default=0,
        )  # materials imported from YAML may have arbitrary numbering

        # Add sets from couplings and boundary conditions to a temp container
        mesh.unlink_nodes()
        mesh_sets = mesh.get_unique_geometry_sets(
            geometry_set_start_indices=start_indices_geometry_set
        )

        # Assign global indices
        #   Nodes
        if len(mesh.nodes) != len(set(mesh.nodes)):
            raise ValueError("Nodes are not unique!")
        for i, node in enumerate(mesh.nodes, start=start_index_nodes):
            node.i_global = i

        #   Elements
        if len(mesh.elements) != len(set(mesh.elements)):
            raise ValueError("Elements are not unique!")
        i = start_index_elements
        nurbs_count = 0

        for element in mesh.elements:
            element.i_global = i
            if isinstance(element, _NURBSPatch):
                element.i_nurbs_patch = nurbs_count
                i += element.get_number_of_elements()
                nurbs_count += 1
                continue
            i += 1

        #   Materials: Get a list of all materials in the mesh,
        #   including nested sub-materials.
        all_materials = [
            material
            for mesh_material in mesh.materials
            for material in _get_all_contained_materials(mesh_material)
        ]
        if len(all_materials) != len(set(all_materials)):
            raise ValueError("Materials are not unique!")
        for i, material in enumerate(all_materials, start=start_index_materials):
            material.i_global = i

        #   Functions
        if len(mesh.functions) != len(set(mesh.functions)):
            raise ValueError("Functions are not unique!")
        for i, function in enumerate(mesh.functions, start=start_index_functions):
            function.i_global = i

        # Dump mesh to input file
        def _dump(section_name: str, items: _List) -> None:
            """Dump list of items to a section in the input file.

            Args:
                section_name: Name of the section
                items: List of items to be dumped
            """
            if not items:  # do not write empty sections
                return
            dumped: list[_Any] = []
            for item in items:
                _dump_item_to_list(dumped, item)

            # Go through FourCIPP to convert to native types
            # TODO this can be simplified/removed by using an internal type converter
            if section_name in self.sections:
                existing = self.pop(section_name)
                existing.extend(dumped)
                dumped = existing

            self.add({section_name: dumped})

        #   Materials
        _dump("MATERIALS", all_materials)

        #   Functions
        for function in mesh.functions:
            self.add({f"FUNCT{function.i_global + 1}": function.data})

        #   Couplings
        #     If there are couplings in the mesh, set the link between the nodes
        #     and elements, so the couplings can decide which DOFs they couple,
        #     depending on the type of the connected beam element.
        if any(
            mesh.boundary_conditions.get((key, _bme.geo.point), [])
            for key in (_bme.bc.point_coupling, _bme.bc.point_coupling_penalty)
        ):
            mesh.set_node_links()

        #   Boundary conditions
        for (bc_key, geometry_key), bc_list in mesh.boundary_conditions.items():
            if bc_list:
                section = (
                    bc_key
                    if isinstance(bc_key, str)
                    else _INPUT_FILE_MAPPINGS["boundary_conditions"][
                        (bc_key, geometry_key)
                    ]
                )
                _dump(section, bc_list)

        #   Additional element sections (NURBS etc.)
        for element in mesh.elements:
            _dump_item_to_section(self, element)

        #   Geometry sets
        for geometry_key, items in mesh_sets.items():
            _dump(
                _INPUT_FILE_MAPPINGS["geometry_sets_geometry_to_condition_name"][
                    geometry_key
                ],
                items,
            )

        #   Nodes
        _dump("NODE COORDS", mesh.nodes)
        #   Elements
        _dump("STRUCTURE ELEMENTS", mesh.elements)

        # TODO: reset all links and counters set in this method.

    def _get_header(self) -> dict:
        """Return the information header for the current BeamMe run.

        Returns:
            A dictionary with the header information.
        """

        header: dict = {"BeamMe": {}}

        header["BeamMe"]["creation_date"] = _datetime.now().isoformat(
            sep=" ", timespec="seconds"
        )

        # application which created the input file
        application_path = _get_application_path()
        if application_path is not None:
            header["BeamMe"]["Application"] = {"path": str(application_path)}

            application_git_sha, application_git_date = _get_git_data(
                application_path.parent
            )
            if application_git_sha is not None and application_git_date is not None:
                header["BeamMe"]["Application"].update(
                    {
                        "git_sha": application_git_sha,
                        "git_date": application_git_date,
                    }
                )

        # BeamMe information
        beamme_git_sha, beamme_git_date = _get_git_data(
            _Path(__file__).resolve().parent
        )
        if beamme_git_sha is not None and beamme_git_date is not None:
            header["BeamMe"]["BeamMe"] = {
                "git_SHA": beamme_git_sha,
                "git_date": beamme_git_date,
            }

        # CubitPy information
        if _cubitpy_is_available():
            cubitpy_git_sha, cubitpy_git_date = _get_git_data(
                _os.path.dirname(_cubitpy.__file__)
            )

            if cubitpy_git_sha is not None and cubitpy_git_date is not None:
                header["BeamMe"]["CubitPy"] = {
                    "git_SHA": cubitpy_git_sha,
                    "git_date": cubitpy_git_date,
                }

        return header

    def _get_application_script(self, application_path: _Path) -> list[str]:
        """Get the script that created this input file.

        Args:
            application_path: Path to the script that created this input file.
        Returns:
            A list of strings with the script that created this input file.
        """

        application_script_lines = [
            "# Application script which created this input file:\n"
        ]

        with open(application_path) as script_file:
            application_script_lines.extend("# " + line for line in script_file)

        return application_script_lines



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Define a Cosserat curve object that can be used to describe warping of
curve-like objects."""

from pathlib import Path as _Path
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from xml.etree import ElementTree as _ET  # nosec B405

import numpy as _np
import pyvista as _pv
import quaternion as _quaternion
from numpy.typing import NDArray as _NDArray
from scipy import integrate as _integrate
from scipy import interpolate as _interpolate
from scipy import optimize as _optimize

from beamme.core.conf import bme as _bme
from beamme.core.rotation import Rotation as _Rotation
from beamme.core.rotation import rotate_coordinates as _rotate_coordinates
from beamme.core.rotation import smallest_rotation as _smallest_rotation


def get_piecewise_linear_arc_length_along_points(
    coordinates: _np.ndarray,
) -> _np.ndarray:
    """Return the accumulated distance between the points.

    Args
    ----
    coordinates:
        Array containing the point coordinates
    """

    n_points = len(coordinates)
    point_distance = _np.linalg.norm(coordinates[1:] - coordinates[:-1], axis=1)
    point_arc_length = _np.zeros(n_points)
    for i in range(1, n_points):
        point_arc_length[i] = point_arc_length[i - 1] + point_distance[i - 1]
    return point_arc_length


def get_spline_interpolation(
    coordinates: _np.ndarray, point_arc_length: _np.ndarray
) -> _interpolate.BSpline:
    """Get a spline interpolation of the given points.

    Args
    ----
    coordinates:
        Array containing the point coordinates
    point_arc_length:
        Arc length for each coordinate

    Return
    ----
    centerline_interpolation:
        The spline interpolation object
    """

    # Interpolate coordinates along arc length
    # Note: The numeric evaluation of the spline interpolation can depend on the
    #     operating system, thus introducing slight numerical differences (~1e-12).
    centerline_interpolation = _interpolate.make_interp_spline(
        point_arc_length, coordinates
    )
    return centerline_interpolation


def get_quaternions_along_curve(
    centerline: _interpolate.BSpline, point_arc_length: _np.ndarray
) -> _NDArray[_quaternion.quaternion]:
    """Get the quaternions along the curve based on smallest rotation mappings.

    The initial rotation will be calculated based on the largest projection of the initial tangent
    onto the cartesian basis vectors.

    Args
    ----
    centerline:
        A function that returns the centerline position for a parameter coordinate t
    point_arc_length:
        Array of parameter coordinates for which the quaternions should be calculated
    """

    centerline_interpolation_derivative = centerline.derivative()

    def basis(i):
        """Return the i-th Cartesian basis vector."""
        basis = _np.zeros([3])
        basis[i] = 1.0
        return basis

    # Get the reference rotation
    t0 = centerline_interpolation_derivative(point_arc_length[0])
    min_projection = _np.argmin(_np.abs([_np.dot(basis(i), t0) for i in range(3)]))
    last_rotation = _Rotation.from_basis(t0, basis(min_projection))

    # Get the rotation vectors along the curve. They are calculated with smallest rotation mappings.
    n_points = len(point_arc_length)
    quaternions = _np.zeros(n_points, dtype=_quaternion.quaternion)
    quaternions[0] = last_rotation.q
    for i in range(1, n_points):
        rotation = _smallest_rotation(
            last_rotation,
            centerline_interpolation_derivative(point_arc_length[i]),
        )
        quaternions[i] = rotation.q
        last_rotation = rotation
    return quaternions


def get_relative_distance_and_rotations(
    coordinates: _np.ndarray, quaternions: _NDArray[_quaternion.quaternion]
) -> _Tuple[
    _np.ndarray, _NDArray[_quaternion.quaternion], _NDArray[_quaternion.quaternion]
]:
    """Get relative distances and rotations that can be used to evaluate
    "intermediate" states of the Cosserat curve."""

    n_points = len(coordinates)
    relative_distances = _np.zeros(n_points - 1)
    relative_distances_rotation = _np.zeros(n_points - 1, dtype=_quaternion.quaternion)
    relative_rotations = _np.zeros(n_points - 1, dtype=_quaternion.quaternion)

    for i_segment in range(n_points - 1):
        relative_distance = coordinates[i_segment + 1] - coordinates[i_segment]
        relative_distance_local = _quaternion.rotate_vectors(
            quaternions[i_segment].conjugate(), relative_distance
        )
        relative_distances[i_segment] = _np.linalg.norm(relative_distance_local)

        smallest_relative_rotation_onto_distance = _smallest_rotation(
            _Rotation(),
            relative_distance_local,
        )
        relative_distances_rotation[i_segment] = (
            smallest_relative_rotation_onto_distance.get_numpy_quaternion()
        )

        relative_rotations[i_segment] = (
            quaternions[i_segment].conjugate() * quaternions[i_segment + 1]
        )

    return relative_distances, relative_distances_rotation, relative_rotations


class CosseratCurve(object):
    """Represent a Cosserat curve in space."""

    def __init__(
        self,
        point_coordinates: _np.ndarray,
        *,
        starting_triad_guess: _Optional[_Rotation] = None,
    ):
        """Initialize the Cosserat curve based on points in 3D space.

        Args:
            point_coordinates: Array containing the point coordinates
            starting_triad_guess: Optional initial guess for the starting triad.
                If provided, this introduces a constant twist angle along the curve.
                The twist angle is computed between:
                - The given starting guess triad, and
                - The automatically calculated triad, rotated onto the first basis vector
                  of the starting guess triad using the smallest rotation.
        """

        self.coordinates = point_coordinates.copy()
        self.n_points = len(self.coordinates)

        # Interpolate coordinates along piece wise linear arc length
        point_arc_length_piecewise_linear = (
            get_piecewise_linear_arc_length_along_points(self.coordinates)
        )
        centerline_interpolation_piecewise_linear = get_spline_interpolation(
            self.coordinates, point_arc_length_piecewise_linear
        )
        centerline_interpolation_piecewise_linear_p = (
            centerline_interpolation_piecewise_linear.derivative(1)
        )

        def ds(t):
            """Arc length along interpolated spline."""
            return _np.linalg.norm(centerline_interpolation_piecewise_linear_p(t))

        # Integrate the arc length along the interpolated centerline, this will result
        # in a more accurate centerline arc length
        self.point_arc_length = _np.zeros(self.n_points)
        for i in range(len(point_arc_length_piecewise_linear) - 1):
            self.point_arc_length[i + 1] = (
                self.point_arc_length[i]
                + _integrate.quad(
                    ds,
                    point_arc_length_piecewise_linear[i],
                    point_arc_length_piecewise_linear[i + 1],
                )[0]
            )

        # Set the interpolation of the (positional) centerline
        self.set_centerline_interpolation()

        # Get the quaternions along the centerline based on smallest rotation mappings
        self.quaternions = get_quaternions_along_curve(
            self.centerline_interpolation, self.point_arc_length
        )

        # Get the relative quantities used to warp the curve
        (
            self.relative_distances,
            self.relative_distances_rotation,
            self.relative_rotations,
        ) = get_relative_distance_and_rotations(self.coordinates, self.quaternions)

        # Check if we have to apply a twist for the rotations
        if starting_triad_guess is not None:
            first_rotation = _Rotation.from_quaternion(self.quaternions[0])
            starting_triad_e1 = starting_triad_guess * [1, 0, 0]
            if _np.dot(first_rotation * [1, 0, 0], starting_triad_e1) < 0.5:
                raise ValueError(
                    "The angle between the first basis vectors of the guess triad you"
                    " provided and the automatically calculated one is too large,"
                    " please check your input data."
                )
            smallest_rotation_to_guess_tangent = _smallest_rotation(
                first_rotation, starting_triad_e1
            )
            relative_rotation = (
                smallest_rotation_to_guess_tangent.inv() * starting_triad_guess
            )
            psi = relative_rotation.get_rotation_vector()
            if _np.linalg.norm(psi[1:]) > _bme.eps_quaternion:
                raise ValueError(
                    "The twist angle can not be extracted as the relative rotation is not plane!"
                )
            twist_angle = psi[0]
            self.twist(twist_angle)

    def set_centerline_interpolation(self):
        """Set the interpolation of the centerline based on the coordinates and
        arc length stored in this object."""
        self.centerline_interpolation = get_spline_interpolation(
            self.coordinates, self.point_arc_length
        )

    def translate(self, vector):
        """Translate the curve by the given vector."""

        self.coordinates += vector
        self.set_centerline_interpolation()

    def rotate(self, rotation: _Rotation, *, origin=None):
        """Rotate the curve and the quaternions."""

        self.quaternions = rotation.get_numpy_quaternion() * self.quaternions
        self.coordinates = _rotate_coordinates(
            self.coordinates, rotation, origin=origin
        )
        self.set_centerline_interpolation()

    def twist(self, twist_angle: float) -> None:
        """Apply a constant twist rotation along the Cosserat curve.

        Args:
            twist_angle: The rotation angle (in radiants).
        """
        material_twist_rotation = _Rotation(
            [1, 0, 0], twist_angle
        ).get_numpy_quaternion()

        self.quaternions = self.quaternions * material_twist_rotation
        self.relative_distances_rotation = (
            material_twist_rotation.conjugate()
            * self.relative_distances_rotation
            * material_twist_rotation
        )
        self.relative_rotations = (
            material_twist_rotation.conjugate()
            * self.relative_rotations
            * material_twist_rotation
        )

    def get_centerline_position_and_rotation(
        self, arc_length: float, **kwargs
    ) -> _Tuple[_np.ndarray, _NDArray[_quaternion.quaternion]]:
        """Return the position and rotation at a given centerline arc
        length."""
        pos, rot = self.get_centerline_positions_and_rotations([arc_length], **kwargs)
        return pos[0], rot[0]

    def get_centerline_positions_and_rotations(
        self, points_on_arc_length, *, factor=1.0
    ) -> _Tuple[_np.ndarray, _NDArray[_quaternion.quaternion]]:
        """Return the position and rotation at given centerline arc lengths.

        If the points are outside of the valid interval, a linear extrapolation will be
        performed for the displacements and the rotations will be held constant.

        This function also allows to scale the curvature along the curve, allowing for a
        "natural" unwrapping of general curves in 3D. We achieve this by scaling the
        "final" curvature along the beam and then evaluating the curve that follows this
        curvature (this would actually require to solve an ODE, but we avoid this by
        using a piecewise constant approximation).

        Args
        ----
        points_on_arc_length: list(float)
            A sorted list with the arc lengths along the curve centerline
        factor: float
            Factor to scale the curvature along the curve.
                factor == 1
                    Use the default positions and the triads obtained via a smallest rotation mapping
                0 <factor < 1
                    Integrate (piecewise constant as evaluated with get_relative_distance_and_rotations)
                    the scaled curvature of the curve to obtain a intuitive wrapping. (factor=0 gives
                    a straight line)
        """

        # Get the points that are within the arc length of the given curve.
        points_on_arc_length = _np.asarray(points_on_arc_length)
        points_in_bounds = _np.logical_and(
            points_on_arc_length > self.point_arc_length[0],
            points_on_arc_length < self.point_arc_length[-1],
        )
        index_in_bound = _np.where(points_in_bounds == True)[0]
        index_out_of_bound = _np.where(points_in_bounds == False)[0]
        points_on_arc_length_in_bound = [
            self.point_arc_length[0],
            *points_on_arc_length[index_in_bound],
            self.point_arc_length[-1],
        ]

        if factor < (1.0 - _bme.eps_quaternion):
            coordinates = _np.zeros_like(self.coordinates)
            quaternions = _np.zeros_like(self.quaternions)
            coordinates[0] = self.coordinates[0]
            quaternions[0] = self.quaternions[0]
            for i_segment in range(self.n_points - 1):
                relative_distance_rotation = _quaternion.slerp_evaluate(
                    _quaternion.quaternion(1),
                    self.relative_distances_rotation[i_segment],
                    factor,
                )
                # In the initial configuration (factor=0) we get a straight curve, so we need
                # to use the arc length here. In the final configuration (factor=1) we want to
                # exactly recover the input points, so we need the piecewise linear distance.
                # Between them, we interpolate.
                relative_distance = (factor * self.relative_distances[i_segment]) + (
                    1.0 - factor
                ) * (
                    self.point_arc_length[i_segment + 1]
                    - self.point_arc_length[i_segment]
                )
                coordinates[i_segment + 1] = (
                    _quaternion.rotate_vectors(
                        quaternions[i_segment] * relative_distance_rotation,
                        [relative_distance, 0, 0],
                    )
                    + coordinates[i_segment]
                )
                quaternions[i_segment + 1] = quaternions[
                    i_segment
                ] * _quaternion.slerp_evaluate(
                    _quaternion.quaternion(1),
                    self.relative_rotations[i_segment],
                    factor,
                )
        else:
            coordinates = self.coordinates
            quaternions = self.quaternions

        sol_r = _np.zeros([len(points_on_arc_length_in_bound), 3])
        sol_q = _np.zeros(
            len(points_on_arc_length_in_bound), dtype=_quaternion.quaternion
        )
        arc_length_spline_interpolation = get_spline_interpolation(
            coordinates, self.point_arc_length
        )
        for i_point, centerline_arc_length in enumerate(points_on_arc_length_in_bound):
            if (
                centerline_arc_length >= self.point_arc_length[0]
                and centerline_arc_length <= self.point_arc_length[-1]
            ):
                for i in range(1, self.n_points):
                    centerline_index = i - 1
                    if self.point_arc_length[i] > centerline_arc_length:
                        break

                # Get the two rotation vectors and arc length values
                arc_lengths = self.point_arc_length[
                    centerline_index : centerline_index + 2
                ]
                q1 = quaternions[centerline_index]
                q2 = quaternions[centerline_index + 1]

                # Linear interpolate the arc length
                xi = (centerline_arc_length - arc_lengths[0]) / (
                    arc_lengths[1] - arc_lengths[0]
                )

                # Perform a spline interpolation for the positions and a slerp
                # interpolation for the rotations
                sol_r[i_point] = arc_length_spline_interpolation(centerline_arc_length)
                sol_q[i_point] = _quaternion.slerp_evaluate(q1, q2, xi)
            else:
                raise ValueError("Centerline value out of bounds")

        # Set the already computed results in the final data structures
        sol_r_final = _np.zeros([len(points_on_arc_length), 3])
        sol_q_final = _np.zeros(len(points_on_arc_length), dtype=_quaternion.quaternion)
        if len(index_in_bound) > 0:
            sol_r_final[index_in_bound] = sol_r[index_in_bound - index_in_bound[0] + 1]
            sol_q_final[index_in_bound] = sol_q[index_in_bound - index_in_bound[0] + 1]

        # Perform the extrapolation at both ends of the curve
        for i in index_out_of_bound:
            arc_length = points_on_arc_length[i]
            if arc_length <= self.point_arc_length[0]:
                index = 0
            elif arc_length >= self.point_arc_length[-1]:
                index = -1
            else:
                raise ValueError("Should not happen")

            length = arc_length - self.point_arc_length[index]
            r = sol_r[index]
            q = sol_q[index]
            sol_r_final[i] = r + _Rotation.from_quaternion(q) * [length, 0, 0]
            sol_q_final[i] = q

        return sol_r_final, sol_q_final

    def project_point(self, p, t0=None) -> float:
        """Project a point to the curve, return the parameter coordinate for
        the projection point."""

        centerline_interpolation_p = self.centerline_interpolation.derivative(1)
        centerline_interpolation_pp = self.centerline_interpolation.derivative(2)

        def f(t):
            """Function to find the root of."""
            r = self.centerline_interpolation(t)
            rp = centerline_interpolation_p(t)
            return _np.dot(r - p, rp)

        def fp(t):
            """Derivative of the Function to find the root of."""
            r = self.centerline_interpolation(t)
            rp = centerline_interpolation_p(t)
            rpp = centerline_interpolation_pp(t)
            return _np.dot(rp, rp) + _np.dot(r - p, rpp)

        if t0 is None:
            t0 = 0.0

        return _optimize.newton(f, t0, fprime=fp)

    def get_pyvista_polyline(self, *, factor: float = 1.0) -> _pv.PolyData:
        """Create a pyvista (vtk) representation of the curve with the
        evaluated triad basis vectors.

        Args:
            factor: Factor to scale the curvature along the curve (see
                `get_centerline_positions_and_rotations` for details).

        Returns:
            A pyvista PolyData object representing the curve.
        """

        positions, rotations = self.get_centerline_positions_and_rotations(
            self.point_arc_length, factor=factor
        )

        poly_line = _pv.PolyData()
        poly_line.points = positions
        cell = _np.arange(0, self.n_points, dtype=int)
        cell = _np.insert(cell, 0, self.n_points)
        poly_line.lines = cell

        rotation_matrices = _quaternion.as_rotation_matrix(rotations)
        for i_dir in range(3):
            poly_line.point_data.set_array(
                rotation_matrices[:, :, i_dir], f"base_vector_{i_dir + 1}"
            )

        return poly_line

    def write_vtk(self, path) -> None:
        """Save a vtk representation of the curve."""
        self.get_pyvista_polyline().save(path)

    def write_pvd_series(
        self,
        pvd_path: _Path | str,
        *,
        factors: list[float] | None = None,
        n_steps: int | None = None,
        binary: bool = True,
    ) -> None:
        """Save a pvd series representing the curve at different states.

        Args:
            pvd_path: Path where to save the pvd file.
            factors: List of factors to scale the curvature along the curve. Mutually exclusive with 'n_steps'.
            n_steps: Number of steps to create a uniform series of factors. Mutually exclusive with 'factors'.
            binary: If True, save the vtk files in binary format.
        """

        pvd_path = _Path(pvd_path)
        if pvd_path.suffix != ".pvd":
            raise ValueError(
                f"The output path must have a .pvd suffix, got {pvd_path.suffix}"
            )

        if factors is not None and n_steps is not None:
            raise ValueError(
                "The keyword arguments 'factors' and 'n_steps' are mutually exclusive."
            )
        if factors is None and n_steps is None:
            raise ValueError(
                "One of the keyword arguments 'factors' or 'n_steps' must be provided."
            )
        if factors is None:
            factors = _np.linspace(0.0, 1.0, num=n_steps)

        pvd_file = _ET.Element("VTKFile", type="Collection", version="0.1")
        collection = _ET.SubElement(pvd_file, "Collection")
        width = max(1, len(str(len(factors) - 1)))
        for i_step, factor in enumerate(factors):
            # TODO: Check if we can use vtp here instead of vtu. Currently this does
            # not work with how we compare files in testing. Since vtu and vtp are
            # basically the same in this case, this solution is fine at the moment.
            factor_file = pvd_path.parent / f"{pvd_path.stem}.{i_step:0{width}d}.vtu"
            _pv.UnstructuredGrid(self.get_pyvista_polyline(factor=factor)).save(
                factor_file, binary=binary
            )
            _ET.SubElement(
                collection,
                "DataSet",
                timestep=str(factor),
                group="",
                part="0",
                file=str(factor_file.relative_to(pvd_path.parent)),
            )

        tree = _ET.ElementTree(pvd_file)
        _ET.indent(tree, space="  ", level=0)
        tree.write(pvd_path, encoding="utf-8", xml_declaration=True)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Functions to create beam meshes along helical paths."""

import warnings as _warnings

import numpy as _np

from beamme.core.mesh import Mesh as _Mesh
from beamme.core.rotation import Rotation as _Rotation

from .beam_line import create_beam_mesh_line as _create_beam_mesh_line


def create_beam_mesh_helix(
    mesh,
    beam_class,
    material,
    axis_vector,
    axis_point,
    start_point,
    *,
    helix_angle=None,
    height_helix=None,
    turns=None,
    warning_straight_line=True,
    **kwargs,
):
    """Generate a helical segment starting at a given start point around a
    predefined axis (defined by axis_vector and axis_point). The helical
    segment is defined by a start_point and exactly two of the basic helical
    quantities [helix_angle, height_helix, turns].

    Args
    ----
    mesh: Mesh
        Mesh that the helical segment will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this segment.
    axis_vector: _np.array, list
        Vector for the orientation of the helical center axis.
    axis_point: _np.array, list
        Point lying on the helical center axis. Does not need to align with
        bottom plane of helix.
    start_point: _np.array, list
        Start point of the helix. Defines the radius.
    helix_angle: float
        Angle of the helix (synonyms in literature: twist angle or pitch
        angle).
    height_helix: float
        Height of helix.
    turns: float
        Number of turns.
    warning_straight_line: bool
        Warn if radius of helix is zero or helix angle is 90 degrees and
        simple line is returned.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    if [helix_angle, height_helix, turns].count(None) != 1:
        raise ValueError(
            "Exactly two arguments of [helix_angle, height_helix, turns]"
            " must be provided!"
        )

    if helix_angle is not None and _np.isclose(_np.sin(helix_angle), 0.0):
        raise ValueError(
            "Helix angle of helix is 0 degrees! "
            + "Change angle for feasible helix geometry!"
        )

    if height_helix is not None and _np.isclose(height_helix, 0.0):
        raise ValueError(
            "Height of helix is 0! Change height for feasible helix geometry!"
        )

    # determine radius of helix
    axis_vector = _np.asarray(axis_vector)
    axis_point = _np.asarray(axis_point)
    start_point = _np.asarray(start_point)

    axis_vector = axis_vector / _np.linalg.norm(axis_vector)
    origin = axis_point + _np.dot(
        _np.dot(start_point - axis_point, axis_vector), axis_vector
    )
    start_point_origin_vec = start_point - origin
    radius = _np.linalg.norm(start_point_origin_vec)

    # create temporary mesh to not alter original mesh
    mesh_temp = _Mesh()

    # return line if radius of helix is 0, helix angle is pi/2 or turns is 0
    if (
        _np.isclose(radius, 0)
        or (helix_angle is not None and _np.isclose(_np.cos(helix_angle), 0.0))
        or (turns is not None and _np.isclose(turns, 0.0))
    ):
        if height_helix is None:
            raise ValueError(
                "Radius of helix is 0, helix angle is 90 degrees or turns is 0! "
                + "Fallback to simple line geometry but height cannot be "
                + "determined based on helix angle and turns! Either switch one "
                + "helix parameter to height of helix or change radius!"
            )

        if warning_straight_line:
            _warnings.warn(
                "Radius of helix is 0, helix angle is 90 degrees or turns is 0! "
                + "Simple line geometry is returned!"
            )

        if helix_angle is not None and height_helix is not None:
            end_point = start_point + height_helix * axis_vector * _np.sign(
                _np.sin(helix_angle)
            )
        elif height_helix is not None and turns is not None:
            end_point = start_point + height_helix * axis_vector

        line_sets = _create_beam_mesh_line(
            mesh_temp,
            beam_class,
            material,
            start_point=start_point,
            end_point=end_point,
            **kwargs,
        )

        # add line to mesh
        mesh.add_mesh(mesh_temp)

        return line_sets

    # generate simple helix
    if helix_angle and height_helix:
        end_point = _np.array(
            [
                radius,
                _np.sign(_np.sin(helix_angle)) * height_helix / _np.tan(helix_angle),
                _np.sign(_np.sin(helix_angle)) * height_helix,
            ]
        )
    elif helix_angle and turns:
        end_point = _np.array(
            [
                radius,
                _np.sign(_np.cos(helix_angle)) * 2 * _np.pi * radius * turns,
                _np.sign(_np.cos(helix_angle))
                * 2
                * _np.pi
                * radius
                * _np.abs(turns)
                * _np.tan(helix_angle),
            ]
        )
    elif height_helix and turns:
        end_point = _np.array(
            [
                radius,
                2 * _np.pi * radius * turns,
                height_helix,
            ]
        )

    helix_sets = _create_beam_mesh_line(
        mesh_temp,
        beam_class,
        material,
        start_point=[radius, 0, 0],
        end_point=end_point,
        **kwargs,
    )

    mesh_temp.wrap_around_cylinder()

    # rotate and translate simple helix to align with necessary axis and starting point
    mesh_temp.rotate(
        _Rotation.from_basis(start_point_origin_vec, axis_vector)
        * _Rotation([1, 0, 0], -_np.pi * 0.5)
    )
    mesh_temp.translate(-mesh_temp.nodes[0].coordinates + start_point)

    # add helix to mesh
    mesh.add_mesh(mesh_temp)

    return helix_sets



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to unit test functions for nodes."""

import numpy as np

from beamme.core.node import Node, NodeCosserat
from beamme.core.rotation import Rotation
from beamme.utils.nodes import (
    adjust_close_nodes,
    get_min_max_coordinates,
    is_node_on_plane,
)


def test_beamme_utils_nodes_adjusting_of_nodes(assert_results_close):
    """Test the mesh function adjust_close_nodes."""

    coordinates = np.array(
        [[0, 0, 0], [1, 0, 0], [2, 0, 0], [1.3, 0, 0], [1.0, 0.3, 0], [2, 0, 0.2]]
    )
    coordinates_averaged = np.array(
        [
            [0, 0, 0],
            [1.1, 0.1, 0],
            [2, 0, 0.1],
            [1.1, 0.1, 0],
            [1.1, 0.1, 0],
            [2, 0, 0.1],
        ]
    )
    nodes = [Node(coordinates=coord) for coord in coordinates]
    adjust_close_nodes(nodes, tol=0.35)
    assert_results_close(coordinates_averaged, [node.coordinates for node in nodes])


def test_beamme_utils_nodes_is_node_on_plane():
    """Test if node on plane function works properly."""

    # node on plane with origin_distance
    node = Node([1.0, 1.0, 1.0])
    assert is_node_on_plane(node, normal=[0.0, 0.0, 1.0], origin_distance=1.0)

    # node on plane with point_on_plane
    node = Node([1.0, 1.0, 1.0])
    assert is_node_on_plane(
        node, normal=[0.0, 0.0, 5.0], point_on_plane=[5.0, 5.0, 1.0]
    )

    # node not on plane with origin_distance
    node = Node([13.5, 14.5, 15.5])
    assert not is_node_on_plane(node, normal=[0.0, 0.0, 1.0], origin_distance=5.0)

    # node not on plane with point_on_plane
    node = Node([13.5, 14.5, 15.5])
    assert not is_node_on_plane(
        node, normal=[0.0, 0.0, 5.0], point_on_plane=[5.0, 5.0, 1.0]
    )


def test_beamme_utils_nodes_get_min_max_coordinates(assert_results_close):
    """Test if the get_min_max_coordinates function works properly."""

    # Create the mesh.
    nodes = []
    nodes.append(Node([0.0, 0.0, 0.0]))
    nodes.append(Node([0.0, 0.0, -1.5]))
    nodes.append(Node([-0.5, 1.0, 3.0]))
    nodes.append(Node([0.5, -1.0, 4.0]))
    nodes.append(NodeCosserat([0.0, 0.0, 0.0], Rotation()))
    nodes.append(NodeCosserat([-0.5, 3.0, 4.0], Rotation()))
    nodes.append(NodeCosserat([2.0, -1.0, 0.0], Rotation()))

    # Check the results.
    min_max = get_min_max_coordinates(nodes)
    ref_solution = [-0.5, -1.0, -1.5, 2.0, 3.0, 4.0]
    assert_results_close(min_max, ref_solution)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test environment utils."""

import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from beamme.utils.environment import (
    cubitpy_is_available,
    get_env_variable,
    get_git_data,
    is_mybinder,
    is_testing,
)


def test_beamme_utils_environment_is_cubitpy_available() -> None:
    """Test is_cubitpy_available function."""

    with patch("beamme.utils.environment._find_spec", return_value=True):
        assert cubitpy_is_available() is True

    with patch("beamme.utils.environment._find_spec", return_value=None):
        assert cubitpy_is_available() is False


def test_beamme_utils_environment_is_mybinder() -> None:
    """Test is_mybinder function."""

    with patch.dict(os.environ, {"BINDER_LAUNCH_HOST": "some_value"}):
        assert is_mybinder() is True

    with patch.dict(os.environ, {}, clear=True):
        assert is_mybinder() is False


def test_beamme_utils_environment_is_testing() -> None:
    """Test is_testing function."""

    with patch.dict(os.environ, {"PYTEST_CURRENT_TEST": "some_value"}):
        assert is_testing() is True

    with patch.dict(os.environ, {}, clear=True):
        assert is_testing() is False


def test_beamme_utils_environment_get_env_variable() -> None:
    """Test get_env_variable function."""

    with patch.dict(os.environ, {"TEST_VAR": "test_value"}):
        assert get_env_variable("TEST_VAR") == "test_value"

    with patch.dict(os.environ, {}, clear=True):
        assert get_env_variable("TEST_VAR", default="default_value") == "default_value"

    with patch.dict(os.environ, {}, clear=True):
        with pytest.raises(
            ValueError, match="Environment variable TEST_VAR is not set"
        ):
            get_env_variable("TEST_VAR")


@patch("beamme.utils.environment._shutil.which")
@patch("beamme.utils.environment._subprocess.run")
def test_beamme_utils_environment_get_git_data_success(mock_run, mock_which):
    """Test get_git_data function with successful git command execution."""

    mock_which.return_value = "/usr/bin/git"

    mock_sha_process = MagicMock()
    mock_sha_process.returncode = 0
    mock_sha_process.stdout = b"abcdef1234567890\n"

    mock_date_process = MagicMock()
    mock_date_process.returncode = 0
    mock_date_process.stdout = b"2024-04-26 12:34:56 +0000\n"

    mock_run.side_effect = [mock_sha_process, mock_date_process]

    sha, date = get_git_data(Path("/path/to/repo"))

    assert sha == "abcdef1234567890"
    assert date == "2024-04-26 12:34:56 +0000"
    assert mock_run.call_count == 2


@patch("beamme.utils.environment._shutil.which")
def test_beamme_utils_environment_get_git_data_git_not_found(mock_which):
    """Test get_git_data function when git executable is not found."""

    mock_which.return_value = None

    with pytest.raises(RuntimeError, match="Git executable not found"):
        get_git_data(Path("/path/to/repo"))


@patch("beamme.utils.environment._shutil.which")
@patch("beamme.utils.environment._subprocess.run")
def test_beamme_utils_environment_get_git_data_subprocess_failure(mock_run, mock_which):
    """Test get_git_data function with subprocess command failure."""

    mock_which.return_value = "/usr/bin/git"

    mock_failed_process = MagicMock()
    mock_failed_process.returncode = 1
    mock_failed_process.stdout = b""

    mock_run.side_effect = [mock_failed_process, mock_failed_process]

    sha, date = get_git_data(Path("/path/to/repo"))

    assert sha is None
    assert date is None
    assert mock_run.call_count == 2
